/*
 *  Copyright (C) 2007 Daniel Bessler <danielb@kiba-dock.org>
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110-1301  USA.
 *
 *  Author : Daniel Bessler <danielb@kiba-dock.org>
*/

#include "kiba-win.h"
#include "kiba-dock.h"
#include "dnd-window.h"

static GtkWidget *dnd_window = NULL;
static KibaObject *dragged = NULL;
static gfloat mouse_dx = 0;
static gfloat mouse_dy = 0;
static gint mouse_x = 0;
static gint mouse_y = 0;

// recalculate win size and change win position, when
// the object changes the needed space
static void
allocate_size(GtkWidget *dnd_window)
{
	gfloat surface_factor_x = 1.0f, surface_factor_y = 1.0f;
	gfloat widget_factor_x = 1.0f, widget_factor_y = 1.0f;
	KibaAllocation allocation = (KibaAllocation) {
		0.0f, 0.0f, dragged->size.width, dragged->size.height
	};
	GtkAllocation alloc;
	
	kiba_object_get_surface_scale(dragged,
		&surface_factor_x, &surface_factor_y);
	kiba_object_get_widget_scale(dragged,
		&widget_factor_x, &widget_factor_y);
	allocation.width *= dragged->cfg_scale_x*widget_factor_x;
	allocation.height *= dragged->cfg_scale_y*widget_factor_y;
	
	allocation.x = mouse_x - mouse_dx*dragged->alloc.width +
		0.5f*(1.0f-surface_factor_x)*allocation.width;
	allocation.y = mouse_y - mouse_dy*dragged->alloc.height +
		0.5f*(1.0f-surface_factor_y)*allocation.height;
	
	allocation.width *= surface_factor_x;
	allocation.height *= surface_factor_y;
	
	alloc.x = (gint) allocation.x;
	alloc.y = (gint) allocation.y;
	alloc.width = (gint) allocation.width;
	alloc.height = (gint) allocation.height;
	
	dnd_window->allocation = alloc;
	gtk_widget_queue_resize_no_redraw(dnd_window);
	gtk_widget_set_size_request(dnd_window,
		alloc.width, alloc.height);
	gtk_window_resize (GTK_WINDOW(dnd_window),
		alloc.width, alloc.height);
	gtk_window_move (GTK_WINDOW(dnd_window),
		alloc.x, alloc.y);
}

static void
area_changed(KibaDrawableObject *drawable_object,
		gpointer data)
{
	allocate_size(data);
}

static gboolean
__expose_event_cb (GtkWidget *widget,
		GdkEventExpose *expose,
		gpointer data)
{
	KibaObject *object = KIBA_OBJECT(data);
	GtkWidget *dnd_widget = GTK_WIDGET(object);
	GtkWidget *child;
	cairo_t *cr;

	cr = gdk_cairo_create (expose->window);
	if (!cr) return TRUE;
	
	// clear background to be transparent
	cairo_set_operator (cr, CAIRO_OPERATOR_SOURCE);
	kiba_transparent_window_draw_background(
		KIBA_TRANSPARENT_WINDOW(widget), cr);

	if (KIBA_IS_DRAWABLE_OBJECT(object)) {
		KibaDrawableObject *drawable_object = KIBA_DRAWABLE_OBJECT(object);
		if (drawable_object->rendered_object->surface) {
			if (drawable_object->alpha > 0.0f) {
				cairo_set_operator (cr, CAIRO_OPERATOR_OVER);
				cairo_set_source_surface(cr,
					drawable_object->rendered_object->surface, 0, 0);
				cairo_paint_with_alpha(cr, drawable_object->alpha);
			}
		}
	} else {
		// FIXME get pixmap from widget or so
	}
	cairo_destroy (cr);

	child = gtk_bin_get_child(GTK_BIN(widget));
	if (child) {
		gtk_container_propagate_expose (
			GTK_CONTAINER (widget), child,  expose);
	 }

	return TRUE;
}

static void
__destroy_cb (GtkWidget *widget,
		gpointer data)
{
	kiba_object_unset_state(dragged,
		KIBA_OBJECT_PRELIGHTED);
	if (KIBA_IS_DRAWABLE_OBJECT(dragged)) {
		kiba_drawable_object_queue_draw(KIBA_DRAWABLE_OBJECT(dragged));
	}
	g_signal_handlers_disconnect_by_func (G_OBJECT(dragged),
		G_CALLBACK(area_changed), dnd_window);
	g_signal_handlers_disconnect_by_func (G_OBJECT(dnd_window),
		G_CALLBACK(__expose_event_cb), dragged);
	g_signal_handlers_disconnect_by_func (G_OBJECT(dnd_window),
		G_CALLBACK(__destroy_cb), dragged);
	dnd_window = NULL;
	dragged = NULL;
}

/// public methods

void
kiba_dnd_window_set_mouse_position(GtkWidget *dnd_window,
		gint x, gint y)
{
	mouse_x = x;
	mouse_y = y;
	allocate_size(dnd_window);
}

GtkWidget*
kiba_dnd_window_new (KibaObject *object,
		gfloat dx, gfloat dy)
{
	GdkScreen *screen = gdk_screen_get_default ();

	g_return_val_if_fail (KIBA_IS_OBJECT(object), dnd_window);
	g_return_val_if_fail (dnd_window == NULL, dnd_window);

	dnd_window = kiba_transparent_window_new ();
	gtk_window_set_decorated (GTK_WINDOW(dnd_window), FALSE);
	gtk_window_set_resizable (GTK_WINDOW(dnd_window), FALSE);
	gtk_window_set_skip_taskbar_hint (GTK_WINDOW(dnd_window), TRUE);
	gtk_window_set_type_hint (GTK_WINDOW(dnd_window),
		GDK_WINDOW_TYPE_HINT_DOCK);
	
	mouse_dx = dx;
	mouse_dy = dy;

	dragged = object;
	kiba_object_set_state(object, KIBA_OBJECT_PRELIGHTED);
	
	// FIXME: fake transparency
	gtk_widget_set_colormap (dnd_window,
		gdk_screen_get_rgba_colormap (screen));

	g_signal_connect (G_OBJECT(object), "area-changed",
		G_CALLBACK(area_changed), dnd_window);
	g_signal_connect_after (G_OBJECT(dnd_window), "expose-event",
		G_CALLBACK(__expose_event_cb), object);
	g_signal_connect (G_OBJECT(dnd_window), "destroy",
		G_CALLBACK(__destroy_cb), object);
	
	allocate_size(dnd_window);

	gtk_window_present (GTK_WINDOW(dnd_window));

	return dnd_window;
}
