/*
 *  Copyright (C) 2006-2007 Daniel Bessler <danielb@kiba-dock.org>
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110-1301  USA.
 *
 *  Author : Daniel Bessler <danielb@kiba-dock.org>
*/

#include "../config.h"
#include <string.h>
#include <math.h>

#include "kiba.h"
#include "kiba-drawable-object.h"
#include "kiba-object.h"
#include "kiba-win.h"
#include "kiba-prefs.h"
#include "kiba-dock.h"
#include "kiba-macros.h"
#include "kiba-screen.h"
#include "kiba-x11.h"

#include "kiba-private.h"

#define KIBA_DRAWABLE_OBJECT_GET_PRIVATE(obj) \
		(G_TYPE_INSTANCE_GET_PRIVATE ((obj), KIBA_TYPE_DRAWABLE_OBJECT, KibaDrawableObjectPrivate))

// rendering the objects as black rectangles
#define OBJECT_SOLID_RECTANGLES FALSE
// background color
#define OBJECT_BACKGROUND 0,0,0,0

G_DEFINE_TYPE (KibaDrawableObject, kiba_drawable_object, KIBA_TYPE_OBJECT)

typedef struct {
	GdkWindow *event_window;
	
	gint fade_id;
	
	gfloat object_alpha; /**< alpha from configuration */
	
	GdkRectangle area;
	
	gboolean queue_update_surface;
	gboolean queue_render_object;
	gboolean queue_update_alpha;
	gboolean queue_update_area;
} KibaDrawableObjectPrivate;


enum {
	SIG_SWITCH_EVENT,
	SIG_QUEUED_DRAW,
	SIG_AREA_CHANGED,
	SIG_ALPHA_CHANGED,
	SIG_KIBA_OBJECT_LAST
};
static guint kiba_drawable_object_signals[SIG_KIBA_OBJECT_LAST] = { 0 };

/// Private Methods

static void
__cfg_scale_changed(GObject *object,
		GParamSpec *pspec,
		gpointer data)
{
	kiba_drawable_object_queue_update_surface(
		KIBA_DRAWABLE_OBJECT(object));
}

static void
__scale_changed(KibaDrawableObject *object,
		gboolean widget_scale_changed,
		gboolean surface_scale_changed,
		gpointer data)
{
	kiba_drawable_object_update_alpha(object);
	kiba_drawable_object_update_area(object);
	kiba_drawable_object_queue_render_object(object);
	kiba_drawable_object_queue_draw(object);
}

static void
__offset_changed(KibaDrawableObject *object,
		gboolean widget_changed,
		gboolean surface_changed,
		gpointer data)
{
	kiba_drawable_object_update_area(object);
	kiba_drawable_object_queue_draw(object);
}

static void
__dragging_toggled(KibaObject *object,
		gboolean dragging, gpointer data)
{
	kiba_drawable_object_update_alpha(
		KIBA_DRAWABLE_OBJECT(object));
}

static void
__destroy_alpha_func(KibaDrawableObject* object,
		gfloat *alpha_ret, gpointer data)
{
	*alpha_ret = KIBA_OBJECT(object)->destroy_factor;
}

static void
__drag_alpha_func(KibaDrawableObject* object,
		gfloat *alpha_ret, gpointer data)
{
	if (kiba_object_get_state(KIBA_OBJECT(object), KIBA_OBJECT_DRAGGING)) {
		*alpha_ret = 0.6f;
	} else {
		*alpha_ret = 1.0f;
	}
}

static void
__object_alpha_func(KibaDrawableObject* object,
		gfloat *alpha_ret, gpointer data)
{
	*alpha_ret = KIBA_DRAWABLE_OBJECT_GET_PRIVATE(object)->object_alpha;
}

static void
__dock_object_size_changed_cb (KibaDock *dock,
		gint size,
		gpointer data)
{
	KibaObject *object = data;
	object->size.width = size;
	object->size.height = size;
	kiba_drawable_object_queue_update_surface(KIBA_DRAWABLE_OBJECT(object));
}

static void
__intersect_areas(KibaAllocation *allocation, GList *features)
{
	GList *l;
	for (l = features; l != NULL; l = l->next) {
		KibaVisualFeatureObject *object = l->data;
		gfloat new_x, new_y;
		new_x = MIN(allocation->x, object->allocation.x);
		new_y = MIN(allocation->y, object->allocation.y);
		allocation->width = MAX(allocation->x + allocation->width,
			object->allocation.x + object->allocation.width) - new_x;
		allocation->height = MAX(allocation->y + allocation->height,
			object->allocation.y + object->allocation.height) - new_y;
		allocation->x = new_x;
		allocation->y = new_y;
	}
}

static void
__render_surface_buffer (KibaDrawableObject *drawable_object, cairo_t *cr)
{
	if (!drawable_object->get_surface) return;
	
	cairo_surface_t *surface = drawable_object->get_surface(drawable_object);
	if (!surface) return;
	
	cairo_set_source_surface(cr, surface, 0, 0);
	cairo_paint(cr);
	cairo_surface_destroy(surface);
}

static void
__paint_features(GList *features,
		cairo_t *cr,
		gfloat scale_x, gfloat scale_y,
		KibaObject *object)
{
	GList *l;
	
	cairo_save(cr);
		// dont scale translations
		cairo_scale(cr, 1.0f/scale_x, 1.0f/scale_y);
		
		for (l = features; l != NULL; l = l->next) {
			KibaVisualFeatureObject *feature_object = l->data;
			KibaAllocation alloc = feature_object->allocation;
			cairo_save(cr);
				cairo_translate (cr, alloc.x, alloc.y);
				
				// feature can be scaled again
				cairo_scale(cr, scale_x, scale_y);
				kiba_visual_feature_draw(
					feature_object->feature, feature_object, cr);
			cairo_restore (cr);
		}
	cairo_restore (cr);
}

gboolean
__scale (KibaDrawableObject *drawable_object, cairo_t *cr)
{
	g_return_if_fail (KIBA_IS_DRAWABLE_OBJECT(drawable_object));
	g_return_if_fail (cr != NULL);
	
	KibaObject *object = KIBA_OBJECT(drawable_object);
	
	if (object->scale_x == 0.0f || object->scale_y == 0.0f) {
		return FALSE;
	}
	cairo_scale (cr, object->scale_x, object->scale_y);

	return TRUE;
}

/// Public methods

void
kiba_drawable_object_update_area (KibaDrawableObject *object)
{
	g_return_if_fail (KIBA_IS_DRAWABLE_OBJECT(object));
	KIBA_DRAWABLE_OBJECT_GET_PRIVATE (object)->queue_update_area = TRUE;
}

/**
 * kiba_drawable_object_update_area:
 * @object: the #KibaDrawableObject instance.
 *
 * Updates the area the objects needs on the window for displaying.
 */
void
kiba_drawable_object_do_update_area (KibaDrawableObject *object)
{
	KibaDrawableObjectPrivate *priv = KIBA_DRAWABLE_OBJECT_GET_PRIVATE (object);
	
	if (!priv->queue_update_area) {
		return;
	}
	priv->queue_update_area = FALSE;
	
	KibaObject *kobj = KIBA_OBJECT(object);
	gfloat width, height;
	const KibaLayoutData *d;
	KibaAllocation area;
	GdkRectangle g_area;
	
	if (!GTK_WIDGET(object)->parent) {
		return;
	}
	
	d = &kiba_layout_data[KIBA_DOCK(GTK_WIDGET(object)->parent)->d.position];
	
	width = ((gfloat) kobj->size.width)*kobj->widget_scale_x*kobj->cfg_scale_x;
	height = ((gfloat) kobj->size.height)*kobj->widget_scale_y*kobj->cfg_scale_y;
	
	area.x = kobj->alloc.x - kobj->surface_offset_x +
		0.5f*width*(1.0f-kobj->surface_scale_x);
	area.y = kobj->alloc.y - kobj->surface_offset_y +
		0.5f*height*(1.0f-kobj->surface_scale_y);
	
	area.height = (kobj->surface_scale_y*height);
	area.width  = (kobj->surface_scale_x*width);
	
	g_area.x = (gint) area.x - 2;
	g_area.y = (gint) area.y - 2;
	g_area.width = (gint) area.width + 4;
	g_area.height = (gint) area.height + 4;
	if (g_area.x != priv->area.x ||
	    g_area.y != priv->area.y ||
	    g_area.width != priv->area.width ||
	    g_area.height != priv->area.height) {
		//kiba_drawable_object_queue_draw (object);
		priv->area = g_area;
		kiba_drawable_object_queue_draw (object);
		
		//g_signal_emit(object,
		//	kiba_drawable_object_signals[SIG_AREA_CHANGED],
		//	0, TRUE);
	}
}

/**
 * kiba_drawable_object_get_area:
 * @object: the #KibaDrawableObject instance.
 */
GdkRectangle
kiba_drawable_object_get_area (KibaDrawableObject *object)
{
	g_return_if_fail (KIBA_IS_DRAWABLE_OBJECT(object));
	return KIBA_DRAWABLE_OBJECT_GET_PRIVATE (object)->area;
}

/**
 * kiba_drawable_object_queue_draw:
 * @object: the #KibaDrawableObject instance.
 *
 * Invalidates the area on the dock used by the object.
 */
void
kiba_drawable_object_queue_draw (KibaDrawableObject *drawable_object)
{
	g_return_if_fail (KIBA_IS_DRAWABLE_OBJECT(drawable_object));
	GtkWidget *widget = GTK_WIDGET(drawable_object);
	KibaDrawableObjectPrivate *priv = KIBA_DRAWABLE_OBJECT_GET_PRIVATE (drawable_object);
	GdkRectangle area = kiba_drawable_object_get_area(drawable_object);
	
	if (widget->parent) {
		kiba_win_invalidate_area(widget->parent->parent, area);
		g_signal_emit(drawable_object,
			kiba_drawable_object_signals[SIG_QUEUED_DRAW],
			0, TRUE);
	}
}

/**
 * kiba_drawable_object_set_alpha:
 * @object: the #KibaDrawableObject instance.
 * @alpha: The alpha of KibaDrawableObject.
 *
 * Sets the alpha factor used for rendering the object surface.
 */
void
kiba_drawable_object_set_alpha (KibaDrawableObject *object,
		gfloat alpha)
{
	g_return_if_fail (KIBA_IS_DRAWABLE_OBJECT(object));
	g_return_if_fail (0.0f <= alpha || alpha <= 1.0f);
	
	KIBA_DRAWABLE_OBJECT_GET_PRIVATE (object)->object_alpha = alpha;
	kiba_drawable_object_update_alpha(object);
	kiba_drawable_object_queue_draw(object);
}

/**
 * kiba_object_event_window:
 * @object: the #KibaDrawableObject instance.
 *
 * Returns: the event catcher window of the object.
 */
GdkWindow*
kiba_drawable_object_event_window(KibaDrawableObject *object)
{
	g_return_val_if_fail(KIBA_IS_DRAWABLE_OBJECT(object), NULL);
	return KIBA_DRAWABLE_OBJECT_GET_PRIVATE (object)->event_window;
}

/**
 * kiba_drawable_object_switch_surface:
 * @object: the #KibaDrawableObject instance.
 *
 * Switches the object's surface, sending the #switch-event signal.
 */
void
kiba_drawable_object_switch_surface (KibaDrawableObject *object)
{
	g_return_if_fail (KIBA_IS_DRAWABLE_OBJECT(object));
	g_signal_emit (object,
		kiba_drawable_object_signals[SIG_SWITCH_EVENT], 0);
}

/**
 * kiba_drawable_object_set_render_object_method:
 * @drawable_object: the #KibaDrawableObject instance.
 * @render_object: #RenderObjectFunc.
 */
void
kiba_drawable_object_set_render_object_method(
		KibaDrawableObject *drawable_object,
		RenderObjectFunc render_object)
{
	g_return_if_fail (KIBA_IS_DRAWABLE_OBJECT(drawable_object));
	drawable_object->render_object = render_object;
}

/**
 * kiba_drawable_object_get_render_object_method:
 * @drawable_object: the #KibaDrawableObject instance.
 */
RenderObjectFunc
kiba_drawable_object_get_render_object_method(
		KibaDrawableObject *drawable_object)
{
	g_return_if_fail (KIBA_IS_DRAWABLE_OBJECT(drawable_object));
	return drawable_object->render_object;
}

/**
 * kiba_drawable_object_queue_update_surface:
 * @drawable_object: the #KibaDrawableObject instance.
 */
void
kiba_drawable_object_queue_update_surface(KibaDrawableObject *drawable_object)
{
	g_return_if_fail(KIBA_IS_DRAWABLE_OBJECT(drawable_object));
	
	KibaDrawableObjectPrivate *priv = KIBA_DRAWABLE_OBJECT_GET_PRIVATE (drawable_object);
	priv->queue_update_surface = TRUE;
	kiba_drawable_object_queue_draw(drawable_object);
}

/**
 * kiba_drawable_object_queue_render_object:
 * @drawable_object: the #KibaDrawableObject instance.
 */
void
kiba_drawable_object_queue_render_object(KibaDrawableObject *drawable_object)
{
	g_return_if_fail(KIBA_IS_DRAWABLE_OBJECT(drawable_object));
	
	KibaDrawableObjectPrivate *priv = KIBA_DRAWABLE_OBJECT_GET_PRIVATE (drawable_object);
	priv->queue_render_object = TRUE;
	kiba_drawable_object_queue_draw(drawable_object);
}

/**
 * kiba_drawable_object_update_surface:
 * @drawable_object: the #KibaDrawableObject instance.
 * @win: the #KibaWin instance.
 */
void
kiba_drawable_object_update_surface(KibaDrawableObject *drawable_object,
		KibaWin *win)
{
	g_return_if_fail(KIBA_IS_DRAWABLE_OBJECT(drawable_object));
	
	KibaDrawableObjectPrivate *priv = KIBA_DRAWABLE_OBJECT_GET_PRIVATE (drawable_object);
	
	if (priv->queue_update_surface) {
		priv->queue_render_object = TRUE;
		priv->queue_update_surface = FALSE;
		drawable_object->update_surface(drawable_object);
	}
	if (priv->queue_render_object) {
		priv->queue_render_object = FALSE;
		KibaObject *object = KIBA_OBJECT(drawable_object);
		GdkRectangle area = kiba_drawable_object_get_area(drawable_object);
		KibaAllocation allocation = (KibaAllocation) {
			0.0f, 0.0f,
			0.0f, 0.0f
		};
		KibaAllocation buf;
		cairo_t *surface_cr, *cr;
		cairo_surface_t *surface;
		
		allocation.width = ((gfloat) object->size.width)*object->widget_scale_x*object->cfg_scale_x;
		allocation.height =((gfloat) object->size.height)*object->widget_scale_y*object->cfg_scale_y;

		allocation.width *= object->surface_scale_x;
		allocation.height *= object->surface_scale_y;
		
		__intersect_areas(&allocation, object->features_morph_front);
		__intersect_areas(&allocation, object->features_morph_back);
		
		buf.width = (gfloat) ((gint) allocation.width)+1.0f;
		buf.height = (gfloat) ((gint) allocation.height)+1.0f;
		gfloat diffw = buf.width-allocation.width;
		gfloat diffh = buf.height-allocation.height;
		allocation.width = buf.width;
		allocation.height = buf.height;
		
		cr = kiba_win_cairo_create(win);
		if (cairo_status (cr) != CAIRO_STATUS_SUCCESS)
			return;
		surface = cairo_surface_create_similar (
					cairo_get_target (cr),
					CAIRO_CONTENT_COLOR_ALPHA,
					(gint) allocation.width,
					(gint) allocation.height);
		cairo_destroy(cr);
		if (cairo_surface_status (surface) != CAIRO_STATUS_SUCCESS)
			return;
		
		surface_cr = cairo_create (surface);
		if (cairo_status (surface_cr) != CAIRO_STATUS_SUCCESS)
			return;
		
		if (drawable_object->rendered_object->surface)
			cairo_surface_destroy(drawable_object->rendered_object->surface);
		
		drawable_object->rendered_object->surface = surface;
		drawable_object->rendered_object->width = allocation.width;
		drawable_object->rendered_object->height = allocation.height;
		
		cairo_set_source_rgba (surface_cr, OBJECT_BACKGROUND);
		cairo_set_operator (surface_cr, CAIRO_OPERATOR_SOURCE);
		cairo_paint (surface_cr);
		
		if (drawable_object->alpha <= 0.0f) {
			cairo_destroy(surface_cr);
			return;
		}
		
		cairo_set_operator (surface_cr, CAIRO_OPERATOR_OVER);
		
		// FIXME: dont know if this makes sense anywhere
		drawable_object->rendered_x = -allocation.x;
		drawable_object->rendered_y = -allocation.y;
		cairo_translate(surface_cr, drawable_object->rendered_x, drawable_object->rendered_y);
		
		drawable_object->rendered_x -= diffw;
		drawable_object->rendered_y -= diffh;
		
		if (__scale (drawable_object, surface_cr)) {
			__paint_features(object->features_morph_back, surface_cr,
					object->surface_scale_x, object->surface_scale_y, object);
		
			cairo_save(surface_cr);
#if OBJECT_SOLID_RECTANGLES == TRUE
			cairo_scale (surface_cr, 1.0f/object->scale_x, 1.0f/object->scale_y);

			cairo_rectangle (surface_cr, 1, 1,
				allocation.width-2, allocation.height-2);
			cairo_set_source_rgba (surface_cr, 0, 0, 0, 1);
			//cairo_fill (surface_cr);
cairo_fill_preserve(surface_cr);
cairo_set_source_rgba (surface_cr, 1, 1, 0, 1);
cairo_stroke(surface_cr);
#else
			drawable_object->render_object (drawable_object, surface_cr);
#endif
			cairo_restore (surface_cr);
			
			__paint_features(object->features_morph_front, surface_cr,
					object->surface_scale_x, object->surface_scale_y, object);
			
			cairo_set_operator (surface_cr, CAIRO_OPERATOR_ATOP);
			__paint_features(object->features_ontop_front, surface_cr,
					object->surface_scale_x, object->surface_scale_y, object);
		}
		
		cairo_destroy(surface_cr);
	}
}

// animation values

void
kiba_drawable_object_update_alpha (KibaDrawableObject *object)
{
	g_return_if_fail (KIBA_IS_DRAWABLE_OBJECT(object));
	KIBA_DRAWABLE_OBJECT_GET_PRIVATE(object)->queue_update_alpha = TRUE;
}

/**
 * kiba_object_update_alpha:
 * @object: the #KibaDrawableObject instance.
 *
 * Applys alpha value changes to the object size.
 *
 * Returns: TRUE if the alpha value changed, otherwise FALSE is returned.
 */
gboolean
kiba_drawable_object_do_update_alpha (KibaDrawableObject *object)
{
	KibaDrawableObjectPrivate *priv = KIBA_DRAWABLE_OBJECT_GET_PRIVATE(object);
	
	if (!priv->queue_update_alpha) {
		return FALSE;
	}
	priv->queue_update_alpha = FALSE;
	    
	gfloat alpha = 1.0f;
	GList *l;

	// add current alpha value of all alpha methods
	for (l = object->alpha_data; l != NULL; l = l->next) {
		KibaDrawableObjectAlphaData *data = l->data;
		gfloat func_alpha = 1.0f;

		data->alpha_func (object, &func_alpha, data->data);

		alpha *= func_alpha;
	}
	
	if (alpha == object->alpha) {
		// value didnt changed
		return FALSE;
	}
	
	object->alpha = alpha;
	kiba_drawable_object_queue_draw (KIBA_DRAWABLE_OBJECT(object));
	
	g_signal_emit(object,
		kiba_drawable_object_signals[SIG_ALPHA_CHANGED],
		0, TRUE);

	return TRUE;
}

/**
 * kiba_drawable_object_remove_alpha_function_by_data:
 * @object: the #KibaDrawableObject instance.
 * @user_data: The user data.
 */
void
kiba_drawable_object_remove_alpha_function_by_data (KibaDrawableObject *object,
		gpointer  user_data)
{
	KibaDrawableObjectAlphaData *data;
	GList *l;
	
	g_return_if_fail(KIBA_IS_DRAWABLE_OBJECT(object));

	for (l = object->alpha_data; l != NULL; l = l->next) {
		data = l->data;
		if (data->data == user_data) {
			g_free (data);
			object->alpha_data = g_list_remove (object->alpha_data, l->data);
			return;
		}
	}

}

/**
 * kiba_drawable_object_add_alpha_function:
 * @object: #KibaDrawableObject instance.
 * @alpha_func: the #KibaDrawableObjectAlphaFunc.
 * @user_data: The user data.
 *
 * Adds alpha function, for alpha animations.
 */
void
kiba_drawable_object_add_alpha_function (KibaDrawableObject *object,
		KibaDrawableObjectAlphaFunc alpha_func,
		gpointer user_data)
{
	KibaDrawableObjectAlphaData *data;
	GList *l;

	for (l = object->alpha_data; l != NULL; l = l->next) {
		data = l->data;
		if (data->alpha_func == alpha_func) {
			data->data = user_data;
			return;
		}
	}

	data = g_new0 (KibaDrawableObjectAlphaData, 1);
	data->alpha_func = alpha_func;
	data->data = user_data;

	object->alpha_data = g_list_append (object->alpha_data, data);
	
	kiba_drawable_object_update_alpha(object);
}

/**
 * kiba_drawable_object_get_alpha:
 * @object: the #KibaDrawableObject instance.
 *
 * Returns: the current alpha.
 */
gfloat
kiba_drawable_object_get_alpha(KibaDrawableObject *object)
{
	g_return_val_if_fail(KIBA_IS_DRAWABLE_OBJECT(object), 1.0f);
	return object->alpha;
}


/// Class Implementations

/**
 * gtk_widget_size_allocate implementation.
 **/
static void
kiba_drawable_object_size_allocate (GtkWidget *widget,
		GtkAllocation *allocation)
{
	KibaDrawableObjectPrivate *priv = KIBA_DRAWABLE_OBJECT_GET_PRIVATE (widget);
	KibaDrawableObject *object = KIBA_DRAWABLE_OBJECT(widget);
	gboolean size_changed = widget->allocation.width != allocation->width ||
	    widget->allocation.height != allocation->height;
	
	if (size_changed ||
	    widget->allocation.y != allocation->y ||
	    widget->allocation.x != allocation->x) {
		widget->allocation = *allocation;
		//kiba_drawable_object_update_area(object);

		if (GDK_IS_WINDOW(priv->event_window))
			gdk_window_move_resize (priv->event_window,
					widget->allocation.x,
					widget->allocation.y,
					widget->allocation.width,
					widget->allocation.height);
	}
	if (size_changed) {
		kiba_drawable_object_queue_render_object(object);
	}
}

/**
 * gtk_widget_unmap implementation.
 **/
static void
kiba_drawable_object_unmap (GtkWidget *widget)
{
	KibaDrawableObjectPrivate *priv = KIBA_DRAWABLE_OBJECT_GET_PRIVATE (widget);
	GtkBin *bin = GTK_BIN (widget);

	GTK_WIDGET_UNSET_FLAGS (widget, GTK_MAPPED);

	gdk_window_hide (priv->event_window);

	if (bin->child && GTK_WIDGET_MAPPED (bin->child))
		gtk_widget_unmap (bin->child);
}

/**
 * gtk_widget_map implementation.
 **/
static void
kiba_drawable_object_map (GtkWidget *widget)
{
	KibaDrawableObjectPrivate *priv = KIBA_DRAWABLE_OBJECT_GET_PRIVATE (widget);
	GtkBin *bin = GTK_BIN (widget);

	GTK_WIDGET_SET_FLAGS (widget, GTK_MAPPED);

	gdk_window_show (priv->event_window);

	if (bin->child && GTK_WIDGET_VISIBLE (bin->child) && !GTK_WIDGET_MAPPED (bin->child))
		gtk_widget_map (bin->child);
}

/**
 * gtk_widget_unrealize implementation.
 **/
static void
kiba_drawable_object_unrealize (GtkWidget *widget)
{
	KibaDrawableObject *object = KIBA_DRAWABLE_OBJECT(widget);
	KibaDrawableObjectPrivate *priv = KIBA_DRAWABLE_OBJECT_GET_PRIVATE (widget);
	GList *l;

	(* GTK_WIDGET_CLASS (kiba_drawable_object_parent_class)->unrealize) (widget);

	gdk_window_set_user_data (priv->event_window, NULL);
	gdk_window_destroy (priv->event_window);
	priv->event_window = NULL;
	
	if (object->rendered_object) {
		if (object->rendered_object->surface)
			cairo_surface_destroy (object->rendered_object->surface);
		g_free (object->rendered_object);
		object->rendered_object = NULL;
	}

	for (l = object->alpha_data; l != NULL; l = object->alpha_data) {
		KibaDrawableObjectAlphaData *data = l->data;
		g_free (data);
		object->alpha_data = g_list_remove (object->alpha_data, l->data);
	}
	g_list_free (object->alpha_data);
	object->alpha_data = NULL;
}

/**
 * gtk_widget_realize implementation.
 **/
static void
kiba_drawable_object_realize (GtkWidget *widget)
{
	KibaDrawableObjectPrivate *priv = KIBA_DRAWABLE_OBJECT_GET_PRIVATE (widget);
	KibaDrawableObject *object = KIBA_DRAWABLE_OBJECT(widget);
	GdkWindowAttr attributes;
	gint attributes_mask;
	gint border_width;
	GtkRequisition requisition;

	GTK_WIDGET_SET_FLAGS (widget, GTK_REALIZED);
	
	object->rendered_object = g_new0 (CairoSurfaceData, 1);
	object->rendered_object->width  = 0;
	object->rendered_object->height = 0;
	object->rendered_object->surface = NULL;

	border_width = GTK_CONTAINER (widget)->border_width;

	gtk_widget_size_request (widget, &requisition);

	if (widget->allocation.x == -1 && widget->allocation.y == -1 &&
	    widget->allocation.width == 1 && widget->allocation.height == 1) {
		GtkAllocation allocation = { 0, 0, 1, 1 };
		if (requisition.width || requisition.height) {
			// non-empty window
			allocation.width = requisition.width;
			allocation.height = requisition.height;
		}
		gtk_widget_size_allocate (widget, &allocation);
	}
	
	widget->window = gtk_widget_get_parent_window (widget);
	g_object_ref (widget->window);
	
	attributes.x = widget->allocation.x;
	attributes.y = widget->allocation.y;
	attributes.width = widget->allocation.width;
	attributes.height = widget->allocation.height;
	attributes.event_mask = gtk_widget_get_events (widget);
	attributes_mask = GDK_WA_X | GDK_WA_Y;

	attributes.window_type = GDK_WINDOW_CHILD;
	attributes.wclass = GDK_INPUT_ONLY;

	priv->event_window = gdk_window_new (widget->window, &attributes, attributes_mask);
	// catching window events on the widget
	gdk_window_set_user_data (priv->event_window, widget);
	// allowing dnd on the window
	gdk_window_register_dnd (priv->event_window);
	gdk_window_set_type_hint (priv->event_window, GDK_WINDOW_TYPE_HINT_DOCK);

	widget->style = gtk_style_attach (widget->style, widget->window);
}

static void
kiba_drawable_object_parent_set(GtkWidget *widget,
		GtkWidget *old_parent)
{
	if (old_parent) {
		// remove old handler
		g_signal_handlers_disconnect_by_func (G_OBJECT(old_parent),
			G_CALLBACK(__dock_object_size_changed_cb), widget);
	}
	if (widget->parent) {
		KibaDock *dock = KIBA_DOCK(widget->parent);
		g_signal_connect(G_OBJECT(widget->parent), "object_size_changed",
			G_CALLBACK(__dock_object_size_changed_cb), widget);
		__dock_object_size_changed_cb (dock, dock->d.object_size, widget);
	}
	
	(* GTK_WIDGET_CLASS (kiba_drawable_object_parent_class)->parent_set) (widget, old_parent);
}

/// Class init

static void
kiba_drawable_object_class_init (KibaDrawableObjectClass *class)
{
	GObjectClass *gobject_class = G_OBJECT_CLASS (class);
	GtkWidgetClass *widget_class = GTK_WIDGET_CLASS (class);

	g_type_class_add_private (class, sizeof (KibaDrawableObjectPrivate));

	// GtkWidget method implementations
	widget_class->realize = kiba_drawable_object_realize;
	widget_class->unrealize = kiba_drawable_object_unrealize;
	widget_class->map = kiba_drawable_object_map;
	widget_class->unmap = kiba_drawable_object_unmap;
	widget_class->size_allocate = kiba_drawable_object_size_allocate;
	widget_class->parent_set = kiba_drawable_object_parent_set;

	// KibaDrawableObject signals
	/**
	 * KibaDrawableObject::switch-event:
	 * @drawable_object: the object on which the signal is emitted
	 *
	 * The switch-event signal is emitted when the object is about
	 * to switch the surface.
	 */
	kiba_drawable_object_signals[SIG_SWITCH_EVENT] =
		g_signal_new ("switch-event",
			G_OBJECT_CLASS_TYPE (class),
			G_SIGNAL_RUN_LAST,
			G_STRUCT_OFFSET (KibaDrawableObjectClass, switch_event),
			NULL, NULL,
			g_cclosure_marshal_VOID__VOID,
			G_TYPE_NONE, 0);
	/**
	 * KibaDrawableObject::alpha-changed:
	 * @drawable_object: the object on which the signal is emitted
	 *
	 * The alpha-changed signal is emitted when a new alpha was calculated
	 * for @drawable_object.
	 */
	kiba_drawable_object_signals[SIG_ALPHA_CHANGED] =
		g_signal_new ("alpha-changed",
			G_OBJECT_CLASS_TYPE (class),
			G_SIGNAL_RUN_LAST,
			G_STRUCT_OFFSET (KibaDrawableObjectClass, alpha_changed),
			NULL, NULL,
			g_cclosure_marshal_VOID__VOID,
			G_TYPE_NONE, 0);
	/**
	 * KibaDrawableObject::area-changed:
	 * @drawable_object: the object on which the signal is emitted
	 *
	 * The area-changed signal is emitted when a new area was calculated
	 * for @drawable_object.
	 * The area defines the area where the object is rendered.
	 */
	kiba_drawable_object_signals[SIG_AREA_CHANGED] =
		g_signal_new ("area-changed",
			G_OBJECT_CLASS_TYPE (class),
			G_SIGNAL_RUN_LAST,
			G_STRUCT_OFFSET (KibaDrawableObjectClass, area_changed),
			NULL, NULL,
			g_cclosure_marshal_VOID__VOID,
			G_TYPE_NONE, 0);
	/**
	 * KibaDrawableObject::queue-draw:
	 * @drawable_object: the object on which the signal is emitted
	 *
	 * The queue-draw signal is emitted when @drawable_object
	 * is about to be redrawn.
	 */
	kiba_drawable_object_signals[SIG_QUEUED_DRAW] =
		g_signal_new ("queued-draw",
			G_OBJECT_CLASS_TYPE (class),
			G_SIGNAL_RUN_LAST,
			G_STRUCT_OFFSET (KibaDrawableObjectClass, queued_draw),
			NULL, NULL,
			g_cclosure_marshal_VOID__VOID,
			G_TYPE_NONE, 0);
}

static void
kiba_drawable_object_init (KibaDrawableObject *object)
{
	KibaDrawableObjectPrivate *priv =
		KIBA_DRAWABLE_OBJECT_GET_PRIVATE (object);
	
	GTK_WIDGET_SET_FLAGS (GTK_WIDGET (object), GTK_NO_WINDOW);
	
	object->alpha_data = NULL;
	object->alpha = 1.0f;
	
	priv->object_alpha = 1.0f;
	priv->queue_update_surface = TRUE;
	priv->queue_render_object = TRUE;
	priv->queue_update_alpha = TRUE;
	priv->queue_update_area = TRUE;
	
	object->update_surface = NULL;
	object->get_surface = NULL;
	object->render_object = __render_surface_buffer;
	object->rendered_object = NULL;
	
	g_signal_connect(G_OBJECT(object), "scale-changed",
		G_CALLBACK(__scale_changed), NULL);
	g_signal_connect(G_OBJECT(object), "offset-changed",
		G_CALLBACK(__offset_changed), NULL);
	g_signal_connect(G_OBJECT(object), "dragging-toggled",
		G_CALLBACK(__dragging_toggled), NULL);
	g_signal_connect (G_OBJECT(object), "notify::scale-x",
		G_CALLBACK (__cfg_scale_changed), NULL);
	g_signal_connect (G_OBJECT(object), "notify::scale-y",
		G_CALLBACK (__cfg_scale_changed), NULL);
	
	kiba_drawable_object_add_alpha_function (object,
		__destroy_alpha_func,
		NULL);
	kiba_drawable_object_add_alpha_function (object,
		__object_alpha_func,
		NULL);
	kiba_drawable_object_add_alpha_function (object,
		__drag_alpha_func,
		NULL);
}

/**
 * kiba_drawable_object_new:
 * @object_id: object id.
 * @scale_x: scale factor in x direction.
 * @scale_y: scale factor in y direction.
 *
 * Creates and returns #KibaDrawableObject container.
 *
 * Returns: #KibaDrawableObject instance created.
 */
GtkWidget*
kiba_drawable_object_new (const gchar *object_id,
		gfloat scale_x, gfloat scale_y)
{
	return g_object_new (KIBA_TYPE_DRAWABLE_OBJECT,
                                 "object_id", object_id,
                                 "scale_x", scale_x,
                                 "scale_y", scale_y, NULL);
}
