
/*
 *  Copyright (C) 2009 Daniel Bessler <danielb@kiba-dock.org>
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110-1301  USA.
 *
 *  Author : Daniel Bessler <daniel@orgizm.net>
*/

#include <cairo.h>
#include <math.h>

#include "kiba.h"
#include "kiba-drawable-object.h"
#include "kiba-prefs.h"
#include "kiba-dock.h"
#include "kiba-private.h"
#include "reflection.h"

G_DEFINE_TYPE (KibaReflectionFeature, kiba_reflection_feature, KIBA_TYPE_VISUAL_FEATURE)

/*** Private methods */

static void
__drag_window_size_allocate_cb (GtkWidget *widget,
		GtkAllocation *allocation,
		gpointer data)
{
	KibaVisualFeatureObject *object = data;
	kiba_visual_feature_queue_resize(object->feature, object);
	gtk_widget_queue_resize_no_redraw(
		GTK_WIDGET(object->object)->parent);
}

static void
__dragging_toggled_cb (KibaObject *object,
		gboolean dragging,
		gpointer data)
{
	KibaDock *dock = KIBA_DOCK(GTK_WIDGET(object)->parent);
	KibaVisualFeature *feature = data;
	GtkWidget *dragging_window;
	KibaVisualFeatureObject *vobject;

	dragging_window = kiba_dock_get_dragging_window(dock);
	
	GList *l;
	for (l = feature->objects; l != NULL; l = l->next) {
		vobject = l->data;
		if (vobject->object == object) break;
	}
	if (dragging) {
		if (dragging_window) {
			g_signal_connect(G_OBJECT(dragging_window), "size-allocate",
				G_CALLBACK(__drag_window_size_allocate_cb), vobject);
		}
	} else {
		if (dragging_window) {
			g_signal_handlers_disconnect_by_func (G_OBJECT(dragging_window),
				G_CALLBACK(__drag_window_size_allocate_cb), vobject);
		}
		kiba_visual_feature_queue_resize(feature, vobject);
		gtk_widget_queue_resize_no_redraw(GTK_WIDGET(dock));
	}
}

static void
__object_queued_draw_cb(KibaObject *object,
		gpointer data)
{
	KibaVisualFeature *feature = data;
	KibaVisualFeatureObject *vobject;
	
	GList *l;
	for (l = feature->objects; l != NULL; l = l->next) {
		vobject = l->data;
		if (vobject->object == object) break;
	}
	
	kiba_visual_feature_queue_draw(feature, vobject);
}

static void
resize_feature_object(
		KibaVisualFeature *feature,
		KibaObject *object)
{
	
	GList *l;
	KibaVisualFeatureObject *vobject;
	for (l = feature->objects; l != NULL; l = l->next) {
		vobject = l->data;
		if (vobject->object == object) break;
	}
	
	kiba_visual_feature_queue_resize(feature, vobject);
}

static void
__scale_changed_cb(KibaObject *object,
		gboolean widget_changed,
		gboolean surface_changed,
		gpointer data)
{
	resize_feature_object(data, object);
}

static void
__offset_changed_cb(KibaObject *object,
		gboolean widget_changed,
		gboolean surface_changed,
		gpointer data)
{
	resize_feature_object(data, object);
}


static void
__dock_position_changed_cb(KibaDock *dock,
	KibaPosition position, gpointer data)
{
	kiba_visual_feature_queue_resize_all(
		(KibaVisualFeature*) data);
}
static cairo_pattern_t*
__reflection_pattern(
		cairo_t *cr,
		const KibaLayoutData *d,
		gfloat alpha,
		gfloat width,
		gfloat height)
{
	cairo_pattern_t *mask;
	if (d->orientation == GTK_ORIENTATION_HORIZONTAL) {
		cairo_scale (cr, 1.0f, -1.0f);
		cairo_translate(cr, 0.0f, -height);
		mask = cairo_pattern_create_linear (
			0.0f,
			(gdouble) ((d->valignment-2)/-2)*height,
			0.0f,
			(gdouble) (d->valignment/2)*height);
	} else {
		cairo_scale (cr, -1.0f, 1.0f);
		cairo_translate(cr, -width, 0.0f);
		mask = cairo_pattern_create_linear (
		(gdouble) ((d->halignment-2)/-2)*width,
			0.0f,
			(gdouble) (d->halignment/2)*width,
			0.0f);
	}
		
	// create linear gradient as mask-pattern to fade out the source
	cairo_pattern_add_color_stop_rgba (mask, 0.0f, 0.0f, 0.0f, 0.0f, 0.0f);
	cairo_pattern_add_color_stop_rgba (mask, 0.3f, 0.0f, 0.0f, 0.0f, alpha*0.01f);
	cairo_pattern_add_color_stop_rgba (mask, 0.6f, 0.0f, 0.0f, 0.0f, alpha*0.125f);
	cairo_pattern_add_color_stop_rgba (mask, 0.85f, 0.0f, 0.0f, 0.0f, alpha*0.5f);
	cairo_pattern_add_color_stop_rgba (mask, 1.0f, 0.0f, 0.0f, 0.0f, alpha*1.0f);
	
	return mask;
}

/// Class implementation

static void
kiba_reflection_feature_object_added(KibaVisualFeature *feature, KibaObject *object)
{
	if (!KIBA_IS_DRAWABLE_OBJECT(object)) {
		return;
	}
	kiba_visual_feature_add_object(feature, object);
	g_signal_connect(G_OBJECT(object), "scale-changed",
		G_CALLBACK(__scale_changed_cb), feature);
	g_signal_connect(G_OBJECT(object), "offset-changed",
		G_CALLBACK(__offset_changed_cb), feature);
	g_signal_connect_after(G_OBJECT(object), "dragging-toggled",
		G_CALLBACK(__dragging_toggled_cb), feature);
	g_signal_connect(G_OBJECT(object), "queued-draw",
		G_CALLBACK(__object_queued_draw_cb), feature);
}

static void
kiba_reflection_feature_object_removed(KibaVisualFeature *feature, KibaObject *object)
{
	kiba_visual_feature_remove_object(feature, object);
	g_signal_handlers_disconnect_by_func (G_OBJECT(object),
		G_CALLBACK(__dragging_toggled_cb), feature);
	g_signal_handlers_disconnect_by_func (G_OBJECT(object),
		G_CALLBACK(__object_queued_draw_cb), feature);
	g_signal_handlers_disconnect_by_func (G_OBJECT(object),
		G_CALLBACK(__offset_changed_cb), feature);
	g_signal_handlers_disconnect_by_func (G_OBJECT(object),
		G_CALLBACK(__scale_changed_cb), feature);
}

static void
kiba_reflection_feature_draw(KibaVisualFeature *feature,
		KibaVisualFeatureObject *object,
		cairo_t *cr)
{
	KibaReflectionFeature *reflection = KIBA_REFLECTION_FEATURE(feature);
	KibaDrawableObject *drawable_object = KIBA_DRAWABLE_OBJECT(object->object);
	cairo_surface_t *surface = drawable_object->rendered_object->surface;
	KibaDock *dock = kiba_visual_feature_get_dock(feature);
	cairo_pattern_t *mask;
	float scale_x, scale_y;
	float height, width;
	const KibaLayoutData *d;
	
	kiba_visual_feature_queue_resize(feature, object);
	
	if (!surface) {
		return;
	}
	
	d = &kiba_layout_data[dock->d.position];
	
	scale_x = object->object->destroy_factor*object->object->surface_scale_x;
	scale_y = object->object->destroy_factor*object->object->surface_scale_y;
	
	width = scale_x*object->object->alloc.width;
	height = scale_y*object->object->alloc.height;
	
	cairo_save(cr);
		// get the alpha mask pattern and mirror surfaces
		mask = __reflection_pattern(cr, d, drawable_object->alpha, width, height);
		// set object surface
		cairo_set_source_surface(cr,
			surface,
			0, 0);
		// apply and draw the current source with the mask-pattern
		cairo_mask (cr, mask);
		// get rid of the mask-pattern again
		cairo_pattern_destroy (mask);
	cairo_restore(cr);
}

static KibaAllocation
kiba_reflection_feature_get_allocation(KibaVisualFeature *feature,
		KibaVisualFeatureObject *object)
{
	KibaDrawableObject *drawable_object = KIBA_DRAWABLE_OBJECT(object->object);
	KibaDock *dock = kiba_visual_feature_get_dock(feature);
	KibaAllocation alloc = (KibaAllocation) {
		0.0f, 0.0f,
		(float) object->object->size.width,
		(float) object->object->size.height
	};
	float offset_x = 0.0f, offset_y = 0.0f;
	float unscaled_w, unscaled_h;
	const KibaLayoutData *d;
	
	d = &kiba_layout_data[dock->d.position];
	
	alloc.width *= object->object->widget_scale_x*object->object->cfg_scale_x;
	alloc.height *= object->object->widget_scale_y*object->object->cfg_scale_y;
	
	unscaled_w = alloc.width;
	unscaled_h = alloc.height;
	
	offset_x = object->object->surface_offset_x;
	offset_y = object->object->surface_offset_y;
	alloc.width *= object->object->surface_scale_x;
	alloc.height *= object->object->surface_scale_y;
	
	alloc.x = ((float)d->halignment-1.0f)*unscaled_w + offset_x;
	alloc.y = ((float)d->valignment-1.0f)*unscaled_h + offset_y;
	
	kiba_visual_feature_translate_to_dnd_window(
		feature, object, &alloc, &offset_x, &offset_y
	);
	alloc.x += offset_x;
	alloc.y += offset_y;
	
	if (d->orientation == GTK_ORIENTATION_HORIZONTAL) {
		alloc.y += (1.0f-object->object->destroy_factor)*unscaled_h;
	} else {
		alloc.x += (1.0f-object->object->destroy_factor)*unscaled_w;
	}
	
	alloc.x += 0.5f*(1.0f-object->object->surface_scale_x)*unscaled_w;
	alloc.y += 0.5f*(1.0f-object->object->surface_scale_y)*unscaled_h;
	
	return alloc;
}

static void
kiba_reflection_win_set(KibaVisualFeature *feature, KibaWin *win)
{
	g_signal_connect(G_OBJECT(KIBA_WIN_CHILD(win)), "position-changed",
		G_CALLBACK(__dock_position_changed_cb), feature);
}

static void
kiba_reflection_feature_class_init (KibaReflectionFeatureClass *class)
{
}

static void
kiba_reflection_feature_init (KibaReflectionFeature *reflection)
{
	KibaVisualFeature *feature = KIBA_VISUAL_FEATURE(reflection);
	KibaDock *dock = kiba_visual_feature_get_dock(feature);
	
	// KibaVisualFeature overrides
	feature->get_allocation = kiba_reflection_feature_get_allocation;
	feature->draw = kiba_reflection_feature_draw;
	feature->object_added = kiba_reflection_feature_object_added;
	feature->object_removed = kiba_reflection_feature_object_removed;
	feature->win_set = kiba_reflection_win_set;
}

KibaVisualFeature*
kiba_reflection_feature_new(KibaWin *win)
{
	KibaVisualFeature *feature = g_object_new (
		    KIBA_TYPE_REFLECTION_FEATURE,
		    "name",  REFLECTION_FEATURE_NAME,
		    "type",  KIBA_VISUAL_FEATURE_ONTOP_OF_DOCK,
		    "front", TRUE,
		    NULL);
	kiba_visual_feature_set_win(feature, win);
	
	return feature;
}
