
/*
 *  Copyright (C) 2009 Daniel Bessler <danielb@kiba-dock.org>
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110-1301  USA.
 *
 *  Author : Daniel Bessler <daniel@orgizm.net>
*/

#include <stdlib.h>
#include <cairo.h>
#include <math.h>

#include "kiba.h"
#include "kiba-drawable-object.h"
#include "kiba-prefs.h"
#include "kiba-dock.h"
#include "kiba-screen.h"
#include "shadow.h"
#include "kiba-private.h"

#define SHADOW_DIFF_KEY "shadow-diff"

// light position, relative to screen mid
#define xp 0.0f

G_DEFINE_TYPE (KibaShadowFeature, kiba_shadow_feature, KIBA_TYPE_VISUAL_FEATURE)

/*** Private methods */

static void
__drag_window_size_allocate_cb (GtkWidget *widget,
		GtkAllocation *allocation,
		gpointer data)
{
	KibaVisualFeatureObject *object = data;
	kiba_visual_feature_queue_resize(object->feature, object);
	gtk_widget_queue_resize_no_redraw(
		GTK_WIDGET(object->object)->parent);
}

static void
__dragging_toggled_cb (KibaObject *object,
		gboolean dragging,
		gpointer data)
{
	KibaDock *dock = KIBA_DOCK(GTK_WIDGET(object)->parent);
	KibaVisualFeature *feature = data;
	GtkWidget *dragging_window;
	KibaVisualFeatureObject *vobject;
	GList *l;

	dragging_window = kiba_dock_get_dragging_window(dock);
	
	for (l = feature->objects; l != NULL; l = l->next) {
		vobject = l->data;
		if (vobject->object == object) break;
	}
	if (dragging && dragging_window) {
		g_signal_connect(G_OBJECT(dragging_window), "size-allocate",
			G_CALLBACK(__drag_window_size_allocate_cb), vobject);
	} else {
		if (dragging_window) {
			g_signal_handlers_disconnect_by_func (G_OBJECT(dragging_window),
				G_CALLBACK(__drag_window_size_allocate_cb), vobject);
		}
		kiba_visual_feature_queue_resize(feature, vobject);
		gtk_widget_queue_resize_no_redraw(GTK_WIDGET(dock));
	}
}

static void
__shadow_color_cb(const char *option_id,
		gpointer value,
		gpointer data)
{
	KibaShadowFeature *shadow = KIBA_SHADOW_FEATURE(data);
	kiba_shadow_feature_set_color(shadow,
		kiba_prefs_rgba_from_string ((char*) value));
}

static void
__shadow_light_distance_cb(const char *option_id,
		gpointer value,
		gpointer data)
{
	KibaShadowFeature *shadow = KIBA_SHADOW_FEATURE(data);
	kiba_shadow_feature_set_light_distance(shadow,
		(float)GPOINTER_TO_INT(value));
}

static void
__shadow_light_height_cb(const char *option_id,
		gpointer value,
		gpointer data)
{
	KibaShadowFeature *shadow = KIBA_SHADOW_FEATURE(data);
	kiba_shadow_feature_set_light_height(shadow,
		(float)GPOINTER_TO_INT(value));
}

static void
__win_size_allocate_cb(GtkWidget *widget,
	GtkAllocation *allocation, gpointer data)
{
	KibaVisualFeature *feature = KIBA_VISUAL_FEATURE(data);
	KibaShadowFeature *shadow = KIBA_SHADOW_FEATURE(feature);
	GList *l;
	
	kiba_visual_feature_queue_resize_all(feature);
	for (l = feature->objects; l != NULL; l = l->next) {
		kiba_visual_feature_do_resize(feature,
			(KibaVisualFeatureObject*) l->data);
	}
}

/// Public methods

void
kiba_shadow_feature_set_color(KibaShadowFeature *shadow, KibaRGBA color)
{
	g_return_if_fail(KIBA_IS_SHADOW_FEATURE(shadow));
	shadow->color = color;
	kiba_visual_feature_queue_draw_all(KIBA_VISUAL_FEATURE(shadow));
}

KibaRGBA
kiba_shadow_feature_get_color(KibaShadowFeature *shadow)
{
	g_return_if_fail(KIBA_IS_SHADOW_FEATURE(shadow));
	return shadow->color;
}

void
kiba_shadow_feature_set_light_height(KibaShadowFeature *shadow, float light_height)
{
	g_return_if_fail(KIBA_IS_SHADOW_FEATURE(shadow));
	shadow->yp = light_height;
	kiba_visual_feature_queue_resize_all(KIBA_VISUAL_FEATURE(shadow));
}

float
kiba_shadow_feature_get_light_height(KibaShadowFeature *shadow)
{
	g_return_if_fail(KIBA_IS_SHADOW_FEATURE(shadow));
	return shadow->yp;
}

void
kiba_shadow_feature_set_light_distance(KibaShadowFeature *shadow, float light_distance)
{
	g_return_if_fail(KIBA_IS_SHADOW_FEATURE(shadow));
	shadow->zp = light_distance;
	kiba_visual_feature_queue_resize_all(KIBA_VISUAL_FEATURE(shadow));
}

float
kiba_shadow_feature_get_light_distance(KibaShadowFeature *shadow)
{
	g_return_if_fail(KIBA_IS_SHADOW_FEATURE(shadow));
	return shadow->zp;
}

/// Class implementation

static void
kiba_shadow_feature_object_added(KibaVisualFeature *feature, KibaObject *object)
{
	if (!KIBA_IS_DRAWABLE_OBJECT(object)) {
		return;
	}
	kiba_visual_feature_add_object(feature, object);
	g_signal_connect_after(G_OBJECT(object), "dragging-toggled",
		G_CALLBACK(__dragging_toggled_cb), feature);
}

static void
kiba_shadow_feature_object_removed(KibaVisualFeature *feature, KibaObject *object)
{
	kiba_visual_feature_remove_object(feature, object);
	
	g_signal_handlers_disconnect_by_func (G_OBJECT(object),
		G_CALLBACK(__dragging_toggled_cb), feature);
	
	float *offset;
	offset = g_object_get_data (G_OBJECT(object), SHADOW_DIFF_KEY);
	if (offset) free(offset);
	g_object_set_data (G_OBJECT(object), SHADOW_DIFF_KEY, NULL);
}

static void
kiba_shadow_feature_draw(KibaVisualFeature *feature,
		KibaVisualFeatureObject *object,
		cairo_t *cr)
{
	KibaShadowFeature *shadow = KIBA_SHADOW_FEATURE(feature);
	KibaDrawableObject *drawable_object = KIBA_DRAWABLE_OBJECT(object->object);
	cairo_surface_t *surface = drawable_object->rendered_object->surface;
	KibaDock *dock = kiba_visual_feature_get_dock(feature);
	const KibaLayoutData *d;
	cairo_matrix_t perpective;
	gfloat scale_x, scale_y;
	gfloat scale_w, scale_h;
	gfloat oh, ow;
	gfloat alpha;
	
	if (object->allocation.width <= 0 ||
	    object->allocation.height <= 0) {
		return;
	}
	
	if (!surface) return;
	
	alpha = shadow->color.a*drawable_object->alpha;
	
	if (alpha <= 0.0f) return;
	
	d = &kiba_layout_data[dock->d.position];
	
	scale_x = object->object->destroy_factor*object->object->surface_scale_x;
	scale_y = object->object->destroy_factor*object->object->surface_scale_y;
	
	oh = (float) drawable_object->rendered_object->height;
	ow = (float) drawable_object->rendered_object->width;
	scale_w = object->allocation.width/ow;
	scale_h = object->allocation.height/oh;
	
	float *offset = g_object_get_data (G_OBJECT(drawable_object), SHADOW_DIFF_KEY);
	g_return_if_fail(offset != NULL);
	
	if (d->orientation == GTK_ORIENTATION_VERTICAL) {
		if (*offset > 0.0f) {
			cairo_matrix_init (&perpective,
				scale_w, ((float) (d->halignment-1))*(oh-object->allocation.height)/ow,
				0.0f, 1.0f,
				-((float) (d->halignment-2)/2)*(object->allocation.width-ow),
				((float) (d->halignment/2))*(object->allocation.height-oh));
		} else {
			cairo_matrix_init (&perpective,
				scale_w, ((float) (d->halignment-1))*(object->allocation.height-oh)/ow,
				0.0, 1.0,
				-((float) (d->halignment-2)/2)*(object->allocation.width-ow),
				-((float) (d->halignment-2)/2)*(object->allocation.height-oh));
		}
	} else {
		if (*offset > 0.0f) {
			cairo_matrix_init (&perpective,
				1.0, 0.0,
				((float) (d->valignment-1))*(ow-object->allocation.width)/oh, scale_h,
				((float) d->valignment/2)*(object->allocation.width-ow),
				-((float) (d->valignment-2)/2)*(object->allocation.height-oh));
		} else {
			cairo_matrix_init (&perpective,
				1.0, 0.0,
				((float) (d->valignment-1))*(object->allocation.width-ow)/oh, scale_h,
				-((float) (d->valignment-2)/2)*(object->allocation.width-ow),
				-((float) (d->valignment-2)/2)*(object->allocation.height-oh));
		}
	}
	cairo_transform(cr, &perpective);
	
	cairo_set_source_rgba (cr, shadow->color.r,
		shadow->color.g, shadow->color.b,
		alpha);
	// using alpha mask from the object surface
	cairo_mask_surface (cr, surface, 0, 0);
}

static KibaAllocation
kiba_shadow_feature_get_allocation(KibaVisualFeature *feature,
		KibaVisualFeatureObject *object)
{
	KibaShadowFeature *shadow = KIBA_SHADOW_FEATURE(feature);
	KibaDrawableObject *drawable_object = KIBA_DRAWABLE_OBJECT(object->object);
	KibaDock *dock = kiba_visual_feature_get_dock(feature);
	gfloat oh = object->object->alloc.height;
	gfloat ow = object->object->alloc.width;
	gfloat ox = object->object->alloc.x;
	gfloat oy = object->object->alloc.y;
	gfloat offset_x, offset_y;
	const KibaLayoutData *d;
	KibaAllocation alloc;
	gfloat left = 0.0f, right = 0.0f;
	GtkWidget *parent;
	gfloat alpha, beta;
	
	d = &kiba_layout_data[dock->d.position];
	
	alloc.x = 0.0f;
	alloc.y = 0.0f;
	
	offset_x = 0.0f;//object->object->surface_offset_x;
	offset_y = 0.0f;//object->object->surface_offset_y;
	
	ow *= object->object->surface_scale_x;
	oh *= object->object->surface_scale_y;
	
	if (d->orientation == GTK_ORIENTATION_HORIZONTAL) {
		ox -= KIBA_SCREEN_DEFAULT->geometry.width/2;
	} else {
		oy -= KIBA_SCREEN_DEFAULT->geometry.height/2;
	}
	
	for (parent = GTK_WIDGET(object->object)->parent; 
			parent != NULL; parent = parent->parent) {
		ox += (float) parent->allocation.x;
		oy += (float) parent->allocation.y;
	}
	
	if (d->orientation == GTK_ORIENTATION_HORIZONTAL) {
		oy = KIBA_SCREEN_DEFAULT->geometry.height - oy;
		
		if (ox != 0.0f) {
			alpha = atan((shadow->yp-oh)/ox);
		} else {
			alpha = 0.0f;
		}
		if (ox+ow != 0.0f) {
			beta = atan((shadow->yp-oh)/(ox+ow));
		} else {
			beta = 0.0f;
		}
		alloc.height = -shadow->zp*oh/(oh-shadow->yp);
		alloc.width = ow;
		
		if (ox < xp) {
			left = MAX(0.0f, -oh/tan(alpha));
			alloc.width += left;
			alloc.x -= left;
		}
		if (xp < ox+ow) {
			right = MAX(0.0f, oh/tan(beta));
			alloc.width += right;
		}
		alloc.y = oh-alloc.height;
	} else {
		ox = KIBA_SCREEN_DEFAULT->geometry.width - ox;
		
		if (ox != 0.0f) {
			alpha = atan((shadow->yp-ow)/oy);
		} else {
			alpha = 0.0f;
		}
		if (ox+ow != 0.0f) {
			beta = atan((shadow->yp-ow)/(oy+oh));
		} else {
			beta = 0.0f;
		}
		alloc.width = -shadow->zp*ow/(ow-shadow->yp);
		alloc.height = oh;
		
		if (oy < xp) {
			left = MAX(0.0f, -ow/tan(alpha));
			alloc.height += left;
			alloc.y -= left;
		}
		if (xp < oy+oh) {
			right = MAX(0.0f, ow/tan(beta));
			alloc.height += right;
		}
		alloc.x = ow-alloc.width;
	}
	alloc.y -= 0.5f*(1.0f-object->object->destroy_factor)*alloc.height;
	alloc.x -= 0.5f*(1.0f-object->object->destroy_factor)*alloc.width;
	
	kiba_visual_feature_translate_to_dnd_window(
		feature, object, &alloc, &offset_x, &offset_y
	);
	
	alloc.x += 0.5f*(1.0f-object->object->surface_scale_x*object->object->destroy_factor)*object->object->alloc.width;
	alloc.y += 0.5f*(1.0f-object->object->surface_scale_y*object->object->destroy_factor)*object->object->alloc.height;
	alloc.x -= offset_x;
	alloc.y -= offset_y;
	
	float *offset;
	offset = g_object_get_data (G_OBJECT(drawable_object), SHADOW_DIFF_KEY);
	if (offset) free(offset);
	offset = (float*) malloc(sizeof(float));
	*offset = right-left;
	g_object_set_data (G_OBJECT(drawable_object),
		SHADOW_DIFF_KEY, offset);

	return alloc;
}

static void
kiba_shadow_feature_win_set(KibaVisualFeature *feature, KibaWin *win)
{
	g_signal_connect(G_OBJECT(win), "size-allocate",
		G_CALLBACK(__win_size_allocate_cb), feature);
}

static void
kiba_shadow_feature_dispose(GObject *obj)
{
	KibaVisualFeature *feature = KIBA_VISUAL_FEATURE(obj);
	KibaDock *dock = kiba_visual_feature_get_dock(feature);
	
	(* G_OBJECT_CLASS (kiba_shadow_feature_parent_class)->dispose) (obj);
	
	g_signal_handlers_disconnect_by_func (G_OBJECT(GTK_WIDGET(dock)->parent),
		G_CALLBACK(__win_size_allocate_cb), feature);
}

static void
kiba_shadow_feature_class_init (KibaShadowFeatureClass *class)
{
	GObjectClass *gobject_class = G_OBJECT_CLASS (class);
	
	// GObject overrides
	gobject_class->dispose = kiba_shadow_feature_dispose;
}

static void
kiba_shadow_feature_init (KibaShadowFeature *shadow)
{
	KibaVisualFeature *feature = KIBA_VISUAL_FEATURE(shadow);
	KibaDock *dock = kiba_visual_feature_get_dock(feature);
	
	shadow->color = (KibaRGBA) { 0.0f, 0.0f, 0.0f, 0.5f };
	shadow->yp = 1540.0f;
	shadow->zp = 1300.0f;
	
	// KibaVisualFeature overrides
	feature->get_allocation = kiba_shadow_feature_get_allocation;
	feature->draw = kiba_shadow_feature_draw;
	feature->object_added = kiba_shadow_feature_object_added;
	feature->object_removed = kiba_shadow_feature_object_removed;
	feature->win_set = kiba_shadow_feature_win_set;
}

KibaVisualFeature*
kiba_shadow_feature_new(KibaWin *win, KibaPrefsFile *file)
{
	KibaVisualFeature *feature = g_object_new (
		    KIBA_TYPE_SHADOW_FEATURE,
		    "name",  SHADOW_FEATURE_NAME,
		    "type",  KIBA_VISUAL_FEATURE_ONTOP_OF_DOCK,
		    "front", TRUE,
		    NULL);
	kiba_visual_feature_set_win(feature, win);
	
	// preferences
	kiba_prefs_add_options (file, (PrefMember[]) {
			{ "shadow_color", "#00000088", KIBA_STRING },
			{ "shadow_light_distance", "1300", KIBA_INT },
			{ "shadow_light_height", "1540", KIBA_INT }
		}, 3);
	// FIXME: remove callbacks
	kiba_prefs_add_callback (file, "shadow_color",
		__shadow_color_cb, feature, TRUE);
	kiba_prefs_add_callback (file, "shadow_light_distance",
		__shadow_light_distance_cb, feature, TRUE);
	kiba_prefs_add_callback (file, "shadow_light_height",
		__shadow_light_height_cb, feature, TRUE);
	
	return feature;
}
