/*
 *  Copyright (C) 2008 Daniel Bessler <danielb@kiba-dock.org>
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110-1301  USA.
 *
 *  Author : Daniel Bessler <danielb@kiba-dock.org>
*/

#include <stdlib.h>
#include <string.h>

#include <gtk/gtk.h>

#define SN_API_NOT_YET_FROZEN
#include <libsn/sn.h>

#include "marshallers.h"
#include "startup-notification.h"

#define TIMEOUT_INTERVAL 200

typedef struct {
	SnLauncherContext *context;
	GObject *object;
	const char *id;
}KibaSnLaunchData;

struct _KibaSn {
	GObject object;

	SnDisplay *display;
	SnMonitorContext *monitorContext;
	GList *launchData;
	gint errorTrapDepth;
	gint timeout_id;
	glong processed_time;
	gint watch_count;
};

struct _KibaSnClass {
	GObjectClass  parent_class;

	void (* launch_started) (KibaSn *sn, GObject *data);
	void (* launch_finished) (KibaSn *sn, gboolean success, GObject *data);
};

G_DEFINE_TYPE (KibaSn, kiba_sn, G_TYPE_OBJECT)

// signal ids
enum {
	SIG_LAUNCH_STARTED = 0,
	SIG_LAUNCH_FINISHED,
	SIG_KIBA_SN_LAST
};

static guint kiba_sn_signals[SIG_KIBA_SN_LAST] = { 0 };

/// Private Methods

static void
error_trap_push (SnDisplay *display,
		Display *xdisplay)
		{
	KibaSn *sn = kiba_sn_get();
	sn->errorTrapDepth++;
}

static void
error_trap_pop (SnDisplay *display,
		Display   *xdisplay)
{
	KibaSn *sn = kiba_sn_get();
	if (sn->errorTrapDepth == 0) {
		fprintf (stderr, "Error trap underflow!\n");
		return;
	}

	XSync (xdisplay, False); // get all errors out of the queue
	sn->errorTrapDepth--;
}

static gboolean
notify_timeout(gpointer data)
{
	KibaSn *sn = data;
	Display *xdisplay;
	long tv_sec_last = 0, tv_usec_last = 0;
	GList *l;
	
	xdisplay = sn_display_get_x_display(sn->display);
	sn->processed_time += TIMEOUT_INTERVAL;
	
	while (XPending(xdisplay) > 0) {
		// process pending events
		XEvent xevent;
		
		XNextEvent (xdisplay, &xevent);
		// process the event in libsn, this may invoke monitor_event_func
		sn_display_process_event (sn->display, &xevent);
	}
	
	for (l = sn->launchData; l != NULL; l = l->next) {
		KibaSnLaunchData *launcherData = l->data;
		SnLauncherContext *context = launcherData->context;
		
		sn_launcher_context_get_last_active_time (context,
			&tv_sec_last, &tv_usec_last);
		if (sn->processed_time/1000L - tv_sec_last > 15L) {
			// complete when nothing happened for 15 seconds
			sn_launcher_context_complete(context);
		}
	}
	
	return TRUE;
}

static void
kiba_sn_watch_events(KibaSn *sn)
{
	sn->watch_count++;
	if (sn->timeout_id == 0) {
		struct timeval tv_now;
		
		gettimeofday(&tv_now, NULL);
		sn->processed_time = tv_now.tv_usec;
		sn->timeout_id = g_timeout_add(TIMEOUT_INTERVAL, notify_timeout, sn);
	}
}

static void
kiba_sn_unwatch_events(KibaSn *sn)
{
	sn->watch_count--;
	if (sn->watch_count == 0 && sn->timeout_id != 0) {
		g_source_remove(sn->timeout_id);
		sn->timeout_id = 0;
	}
}

static void
monitor_event_func (SnMonitorEvent *event,
		void *user_data)
{
	KibaSn *sn = kiba_sn_get();
	SnMonitorContext *context;
	SnStartupSequence *sequence;
	SnLauncherContext *launcherContext;
	KibaSnLaunchData *data = NULL;
	const char *id;
	GList *l;

	context = sn_monitor_event_get_context (event);
	sequence = sn_monitor_event_get_startup_sequence (event);

	// get the launch data associated to this event
	id = sn_startup_sequence_get_id (sequence);
	for (l = sn->launchData; l != NULL; l =l->next) {
		KibaSnLaunchData *launcherData = l->data;
		if (g_str_has_suffix (launcherData->id, id)) {
			// FIXME: libstartup-notify may change the id
			data = launcherData;
			break;
		}
	}
	if (data == NULL) {
		return;
	}
	
	switch (sn_monitor_event_get_type (event)) {
		case SN_MONITOR_EVENT_INITIATED: {
			if (G_IS_OBJECT(data->object)) {
				g_signal_emit (sn,
					kiba_sn_signals[SIG_LAUNCH_STARTED],
					0, data->object);
			}
			break;
		}
		case SN_MONITOR_EVENT_COMPLETED:
		case SN_MONITOR_EVENT_CANCELED: {
			gboolean complete = (sn_monitor_event_get_type (event) == SN_MONITOR_EVENT_COMPLETED);
			if (G_IS_OBJECT(data->object)) {
				g_signal_emit (sn,
					kiba_sn_signals[SIG_LAUNCH_FINISHED],
					0, complete, data->object);
			}
			// cleaning up
			kiba_sn_unwatch_events(sn);
			if (g_list_find(sn->launchData, data)) {
				sn_launcher_context_unref (data->context);
				sn->launchData = g_list_remove(sn->launchData, data);
				g_free(data);
			}
			break;
		}
		default:
			break;
	}
}

/// Public Methods

/**
 * kiba_startup_notification_launch:
 * @sn: The KibaSn Instance.
 * @command: The command to monitor.
 * @object: Custom user data to pass to the signal handler.
 *
 * Monitors the next time @command is launched.
 * The KibaSn instance will emit a "launch-finished" signal then.
 */
void
kiba_sn_launch (const char *command,
		GObject *object)
{
	KibaSn *sn = kiba_sn_get();
	Display *xdisplay = sn_display_get_x_display(sn->display);
	KibaSnLaunchData *data = g_new0(KibaSnLaunchData, 1);

	data->object = object;
	data->context = sn_launcher_context_new (sn->display, DefaultScreen (xdisplay));
	sn_launcher_context_initiate (data->context, "kiba-launcher",
		command, CurrentTime);
	// remember the id
	data->id = sn_launcher_context_get_startup_id(data->context);
	sn->launchData = g_list_append(sn->launchData, data);
 	kiba_sn_watch_events(sn);
}

/**
 * kiba_sn_get:
 *
 * Returns the KibaSn instance.
 * You dont own a refernce on it, so dont call g_object_unref().
 *
 * Returns: The KibaSn Instance created.
 */
KibaSn*
kiba_sn_get (void)
{
	static KibaSn *sn = NULL;
	if (!KIBA_IS_SN(sn))
		sn = g_object_new (KIBA_TYPE_SN, NULL);
	return sn;
}

/// custom implementations

static void
kiba_sn_dispose (GObject *object)
{
	KibaSn *sn = KIBA_SN(object);
	GList *l;

	if (sn->timeout_id != 0) {
		g_source_remove (sn->timeout_id);
	}

	for (l = sn->launchData; l != NULL; l = sn->launchData) {
		KibaSnLaunchData *data = l->data;
		sn_launcher_context_unref (data->context);
		sn->launchData = g_list_remove(sn->launchData, data);
		g_free(data);
	}
	g_list_free(sn->launchData);

	sn_monitor_context_unref (sn->monitorContext);
	sn_display_unref (sn->display);
}

/*** Class methods ***/

static void
kiba_sn_class_init (KibaSnClass *class)
{
	GObjectClass *gobject_class = G_OBJECT_CLASS(class);

	gobject_class->dispose = kiba_sn_dispose;

	kiba_sn_signals[SIG_LAUNCH_STARTED] =
		g_signal_new ("launch-started",
			G_OBJECT_CLASS_TYPE (class),
			G_SIGNAL_RUN_LAST,
			G_STRUCT_OFFSET (KibaSnClass, launch_started),
			NULL, NULL,
			g_cclosure_marshal_VOID__OBJECT,
			G_TYPE_NONE, 1,
			G_TYPE_OBJECT);
	kiba_sn_signals[SIG_LAUNCH_FINISHED] =
		g_signal_new ("launch-finished",
			G_OBJECT_CLASS_TYPE (class),
			G_SIGNAL_RUN_LAST,
			G_STRUCT_OFFSET (KibaSnClass, launch_finished),
			NULL, NULL,
			_kiba_marshal_VOID__BOOLEAN_OBJECT,
			G_TYPE_NONE, 2,
			G_TYPE_BOOLEAN, G_TYPE_OBJECT);
}

static void
kiba_sn_init (KibaSn *sn)
{
	Display *xdisplay;
	SnLauncherContext *context;
	SnMonitorContext *monitorContext;

	sn->errorTrapDepth = 0;
	sn->timeout_id = 0;
	sn->processed_time = 0L;
	sn->watch_count = 0;
	sn->display = NULL;
	sn->launchData = NULL;

	xdisplay = XOpenDisplay (NULL);
	if (xdisplay == NULL) {
		fprintf (stderr, "Could not open display\n");
		return;
	}

	if (getenv ("LIBSN_SYNC") != NULL) {
		XSynchronize (xdisplay, True);
	}

	sn->display = sn_display_new (xdisplay,
						error_trap_push,
						error_trap_pop);

	/* We have to select for property events on at least one
	* root window (but not all as INITIATE messages go to
	* all root windows)
	*/
	XSelectInput (xdisplay, DefaultRootWindow (xdisplay),
					PropertyChangeMask);
	sn->monitorContext = sn_monitor_context_new (
		sn->display, DefaultScreen (xdisplay),
		monitor_event_func,
		NULL, NULL);
}
