/*
 *  Copyright (C) 2006-2009 Daniel Bessler <danielb@kiba-dock.org>
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110-1301  USA.
 *
 *  Author : Daniel Bessler <danielb@kiba-dock.org>
*/

#include <gdk/gdk.h>
#include <gdk/gdkx.h>
#include <gdk-pixbuf/gdk-pixbuf.h>
#include <gdk-pixbuf-xlib/gdk-pixbuf-xlib.h>
#include <X11/Xlib.h>
#include <X11/Xatom.h>

#include <kiba-transparent-window.h>
#include <kiba-screen.h>

#define KIBA_TRANSPARENT_WINDOW_GET_PRIVATE(obj) \
		(G_TYPE_INSTANCE_GET_PRIVATE ((obj), \
		KIBA_TYPE_TRANSPARENT_WINDOW, KibaTransparentWindowPrivate))

// TRUE to see only the rerendered parts of the window
#define CLIPPING_CHECK FALSE
// background color
#define WINDOW_BACKGROUND 0,0,0,0

#define ATOM(a) XInternAtom(display, #a, False)

G_DEFINE_TYPE (KibaTransparentWindow, kiba_transparent_window, GTK_TYPE_WINDOW)

typedef struct _KibaTransparentWindowPrivate KibaTransparentWindowPrivate;

struct _KibaTransparentWindowPrivate {
	gboolean composited; // true if composite manager running
};

typedef struct {
	GdkScreen *screen;
	GdkPixbuf *pixbuf;
}DesktopPixbufData;
static GList *desktop_pixbufs = NULL;

static GdkPixbuf*
get_desktop_background (KibaTransparentWindow *win)
{
	GtkWidget *widget = GTK_WIDGET(win);
	Colormap cmap = GDK_COLORMAP_XCOLORMAP(
		gtk_widget_get_colormap(widget));
	Visual *visual = GDK_VISUAL_XVISUAL(
		gtk_widget_get_visual(widget));
	Display *display = GDK_DISPLAY();
	XID root_window = gdk_x11_drawable_get_xid(
		gdk_screen_get_root_window (gtk_widget_get_screen(widget)));
	GdkPixbuf *pixbuf = NULL;
	gint format, i;
	Atom type;
	unsigned long nitems, bytes;
	unsigned int n;
	Pixmap *pixmap = 0;
	
	if (
			(
			 XGetWindowProperty(display, root_window, ATOM(_XROOTPMAP_ID), 0,
				 1, False, ATOM(PIXMAP), &type, &format,
				 &nitems, &bytes, (unsigned char**)&pixmap) == Success ||
			 XGetWindowProperty(display, root_window, ATOM(ESETROOT_PMAP_ID), 0,
				 1, False, ATOM(PIXMAP), &type, &format,
				 &nitems, &bytes, (unsigned char**)&pixmap) == Success
			)
			&&
			nitems == 1L && format == 32) {
		pixbuf = gdk_pixbuf_xlib_get_from_drawable (
			NULL, // destination
			*pixmap,
			cmap, visual, // colormap and visual, ignored for wins
			0, 0, // source offset
			0, 0, // destination offset
			KIBA_SCREEN_DEFAULT->geometry.width,
			KIBA_SCREEN_DEFAULT->geometry.height);
	}
	if (pixmap) {
		XFree(pixmap);
	}
	
	return pixbuf;
}

static void
load_desktop_pixbuf(KibaTransparentWindow *win, GdkScreen *screen)
{
	DesktopPixbufData *pixbuf_data = NULL;
	GList *l;
	
	// try finding previosly loaded pixbuf
	for (l = desktop_pixbufs; l != NULL; l = l->next) {
		DesktopPixbufData *buf = l->data;
		if (buf->screen == screen) {
			pixbuf_data = buf;
			break;
		}
	}
	
	// if none found, create a new one
	if (!pixbuf_data) {
		pixbuf_data = g_new0(DesktopPixbufData, 1);
		pixbuf_data->screen = screen;
		pixbuf_data->pixbuf = get_desktop_background(win);
		desktop_pixbufs = g_list_append(desktop_pixbufs, pixbuf_data);
	}
	
	win->desktop_background = g_object_ref(pixbuf_data->pixbuf);
}

static void
unload_desktop_pixbuf(KibaTransparentWindow *win, GdkScreen *screen)
{
	DesktopPixbufData *pixbuf_data = NULL;
	GList *l;
	
	// clean up desktop pixmap
	win->desktop_background = NULL;
	for (l = desktop_pixbufs; l != NULL; l = l->next) {
		pixbuf_data = l->data;
		if (pixbuf_data->screen == screen) {
			g_object_unref(pixbuf_data->pixbuf);
			if (!G_IS_OBJECT(pixbuf_data->pixbuf)) {
				desktop_pixbufs = g_list_remove(desktop_pixbufs, pixbuf_data);
				g_free(pixbuf_data);
			}
			break;
		}
	}
}

/// public methods

/**
 * kiba_transparent_window_draw_background:
 * @win: the #KibaTransparentWindow instance.
 * @cr: the drawing context.
 * 
 * Renders transparent background on a drawing context.
 */
void
kiba_transparent_window_draw_background (KibaTransparentWindow *win, cairo_t *cr)
{
	KibaTransparentWindowPrivate *priv =
		KIBA_TRANSPARENT_WINDOW_GET_PRIVATE (win);
	GtkWidget *widget = GTK_WIDGET(win);
	
	if (priv->composited) {
#if CLIPPING_CHECK == TRUE
		cairo_set_source_rgba (cr, 0, 0, 0, 1);
#else
		cairo_set_source_rgba (cr, WINDOW_BACKGROUND);
#endif
	} else if (G_IS_OBJECT(win->desktop_background)) {
		// FIXME: position not up to date on window move
		gdk_cairo_set_source_pixbuf (cr, win->desktop_background,
			-widget->allocation.x, -widget->allocation.y);
	} else {
		cairo_set_source_rgba (cr, 1, 1, 1, 1);
	}
	cairo_paint (cr);
}

/**
 * kiba_transparent_window_is_composited:
 * @win: the #KibaTransparentWindow instance.
 *
 * Returns: wether this window is composited.
 */
gboolean
kiba_transparent_window_is_composited (KibaTransparentWindow *win)
{
	return KIBA_TRANSPARENT_WINDOW_GET_PRIVATE (win)->composited;
}

/// signal handler

static void
screen_background_changed(KibaScreen *screen,
		gpointer data)
{
	KibaTransparentWindow *win = data;
	KibaTransparentWindowPrivate *priv =
		KIBA_TRANSPARENT_WINDOW_GET_PRIVATE (win);
	
/*
	if (G_IS_OBJECT(win->desktop_background)) {
		g_object_unref(G_OBJECT(win->desktop_background));
		win->desktop_background = NULL;
	}
	win->desktop_background = get_desktop_background(win);
	if (!priv->composited) {
		gtk_widget_queue_draw(GTK_WIDGET(win));
	}
*/
}

static void
composited_changed(GdkScreen *screen,
		gpointer data)
{
	KibaTransparentWindow *win = data;
	KibaTransparentWindowPrivate *priv =
		KIBA_TRANSPARENT_WINDOW_GET_PRIVATE (win);
	
	priv->composited = gdk_screen_is_composited (screen);
	
	if (!GDK_IS_WINDOW(GTK_WIDGET(win)->window)) {
		return;
	}
	
	gtk_widget_queue_draw(GTK_WIDGET(win));
}

/// Widget Class Implementations

/*
 * gtk_widget_screen_changed implementation.
 */
static void
transparent_window_screen_changed (GtkWidget *widget,
		GdkScreen *previous_screen)
{
	KibaTransparentWindow *win =
		KIBA_TRANSPARENT_WINDOW(widget);
	KibaTransparentWindowPrivate *priv =
		KIBA_TRANSPARENT_WINDOW_GET_PRIVATE (widget);
	GdkScreen *screen = gtk_widget_get_screen (widget);
    
	// keep track of composite manager
	if (GDK_IS_SCREEN(previous_screen)) {
		g_signal_handlers_disconnect_by_func (G_OBJECT(previous_screen),
			G_CALLBACK(composited_changed), widget);
		
		unload_desktop_pixbuf(win, previous_screen);
	}
	g_signal_connect (G_OBJECT(screen), "composited-changed",
		G_CALLBACK(composited_changed), widget);
	load_desktop_pixbuf(win, screen);
	composited_changed(screen, widget);
}

/// Class init

static void
kiba_transparent_window_class_init (KibaTransparentWindowClass *class)
{
	GtkWidgetClass *widget_class = GTK_WIDGET_CLASS (class);
	g_type_class_add_private (class, sizeof (KibaTransparentWindowPrivate));
	widget_class->screen_changed = transparent_window_screen_changed;
}

static void
kiba_transparent_window_init (KibaTransparentWindow *transparent_window)
{
	KibaTransparentWindowPrivate *priv =
		KIBA_TRANSPARENT_WINDOW_GET_PRIVATE (transparent_window);
	
	transparent_window_screen_changed (GTK_WIDGET (transparent_window), NULL);
	// keep track of background image
	g_signal_connect(G_OBJECT(kiba_screen_get()), "background-changed",
		G_CALLBACK(screen_background_changed), transparent_window);
}

/**
 * transparent_window_new:
 *
 * Creates and returns a KibaTransparentWindow Instance.
 *
 * Returns: the KibaTransparentWindow Instance created.
 */
GtkWidget* kiba_transparent_window_new ()
{
	return g_object_new (KIBA_TYPE_TRANSPARENT_WINDOW, NULL);
}
