/*
 *  Copyright (C) 2009 Daniel Bessler <danielb@kiba-dock.org>
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110-1301  USA.
 *
 *  Author : Daniel Bessler <danielb@kiba-dock.org>
*/

#include <stdlib.h>
#include <kiba-visual-feature.h>
#include <kiba-win.h>
#include "kiba-private.h"

#define KIBA_VISUAL_FEATURE_GET_PRIVATE(obj) \
	(G_TYPE_INSTANCE_GET_PRIVATE ((obj), \
	KIBA_TYPE_VISUAL_FEATURE, KibaVisualFeaturePrivate))

G_DEFINE_TYPE (KibaVisualFeature, kiba_visual_feature, G_TYPE_OBJECT)

typedef struct {
	KibaWin *win;
	KibaVisualFeatureType type;
	gboolean front;
	gchar *name;
} KibaVisualFeaturePrivate;

/* property ids */
enum {
	_PROP_NAME= 1,
	_PROP_TYPE,
	_PROP_FRONT
};

static KibaAllocation __get_allocation(KibaVisualFeature *feature,
		KibaVisualFeatureObject *object)
{
	g_return_if_fail(KIBA_IS_VISUAL_FEATURE(feature));
	g_return_if_fail(object != NULL);
	
	object->allocation = feature->get_allocation(feature, object);
	object->abs_allocation.x = object->object->alloc.x -
		object->object->surface_offset_x + object->allocation.x;
	object->abs_allocation.y = object->object->alloc.y -
		object->object->surface_offset_y + object->allocation.y;
	object->abs_allocation.width = object->allocation.width;
	object->abs_allocation.height = object->allocation.height;
	
	return object->allocation;
}

static void __invalidate_kiba_alloc (KibaWin *win,
                              KibaAllocation alloc) {
	GdkRectangle area;
	area.x = (int) alloc.x - 1;
	area.y = (int) alloc.y - 1;
	area.width = (int) alloc.width + 2;
	area.height = (int) alloc.height + 2;
	if (GTK_WIDGET_REALIZED(win))
		kiba_win_invalidate_area(GTK_WIDGET(win), area);
}

static void __destroy_cb(GtkWidget *widget,
                         gpointer data) {
	KibaVisualFeatureObject *object = data;
	kiba_visual_feature_remove_object(object->feature, object->object);
}

static void __object_added(GtkContainer *container,
	GtkWidget *child, gpointer data)
{
	KibaVisualFeature *feature = data;
	kiba_visual_feature_object_added(data, KIBA_OBJECT(child));
}

static void __object_removed(GtkContainer *container,
	GtkWidget *child, gpointer data)
{
	if (KIBA_IS_VISUAL_FEATURE(data))
		kiba_visual_feature_object_removed(KIBA_VISUAL_FEATURE(data), KIBA_OBJECT(child));
}

static void __dock_added_cb(GtkContainer *container,
	    GtkWidget *child, gpointer data)
{
	KibaVisualFeature *feature = data;
	KibaVisualFeaturePrivate *priv = KIBA_VISUAL_FEATURE_GET_PRIVATE (feature);
	KibaDock *dock = KIBA_DOCK(child);
	GList *l, *children;
	
	if (priv->type == KIBA_VISUAL_FEATURE_FIRST) {
		return;
	}
	
	if (feature->win_set)
		feature->win_set(feature, KIBA_WIN(container));
	
	children = gtk_container_get_children(GTK_CONTAINER(dock));
	for (l = children; l != NULL; l = children) {
		GtkWidget* object = l->data;
		children = g_list_remove(children, object);
		__object_added(GTK_CONTAINER(dock), object, feature);
	}
	g_list_free(children);
	
	g_signal_connect_after(G_OBJECT(dock), "add",
		G_CALLBACK(__object_added), feature);
	g_signal_connect_after(G_OBJECT(dock), "remove",
		G_CALLBACK(__object_removed), feature);
	
}

static void __dock_removed_cb(GtkContainer *container,
	    GtkWidget *child, gpointer data)
{
	g_signal_handlers_disconnect_by_func (G_OBJECT(child),
		G_CALLBACK(__object_added), data);
	g_signal_handlers_disconnect_by_func (G_OBJECT(child),
		G_CALLBACK(__object_removed), data);
}

/*** PUBLIC ***/
	
void kiba_visual_feature_translate_to_dnd_window(
		KibaVisualFeature *feature,
		KibaVisualFeatureObject *object,
		KibaAllocation *alloc,
		float *offset_x,
		float *offset_y)
{
	KibaDrawableObject *drawable_object = KIBA_DRAWABLE_OBJECT(object->object);
	GtkWidget *dragging_window, *dragging_child;
	KibaDock *dock = kiba_visual_feature_get_dock(feature);
	const KibaLayoutData *d;
	
	d = &kiba_layout_data[dock->d.position];
	
	dragging_window = kiba_dock_get_dragging_window(dock);
	dragging_child = kiba_dock_get_dragging_child(dock);
	if (dragging_window && dragging_child == GTK_WIDGET(drawable_object)) {
		GtkWidget *win = GTK_WIDGET(dock)->parent;
		float x_pos, y_pos, x_diff, y_diff;
		gint x,y;
		
		// TODO: remmeber mouse position somewhere
		gtk_window_get_position(GTK_WINDOW(dragging_window), &x, &y);
		if (d->orientation == GTK_ORIENTATION_HORIZONTAL) {
			x_pos = win->allocation.x + object->object->alloc.x +
				(d->valignment/2)*(object->object->alloc.width);
			y_pos = win->allocation.y + object->object->alloc.y +
				(d->valignment/2)*(object->object->alloc.height);
			x_diff = x_pos - (dragging_window->allocation.x +
				(d->valignment/2)*(drawable_object->rendered_object->width));
			y_diff = y_pos - (y +
				(d->valignment/2)*(drawable_object->rendered_object->height));

			alloc->x -= x_diff;
			if (d->valignment == 2) {
				if (y_diff > 0) {
					*offset_y += y_diff;
				} else {
					alloc->y -= y_diff;
				}
			} else {
				if (y_diff < 0) {
					*offset_y += y_diff;
				} else {
					alloc->y -= y_diff;
				}
			}
		} else {
			x_pos = win->allocation.x + object->object->alloc.x +
				(d->halignment/2)*(object->object->alloc.width);
			y_pos = win->allocation.y + object->object->alloc.y +
				(d->halignment/2)*(object->object->alloc.height);
			x_diff = x_pos - (x +
				(d->halignment/2)*(drawable_object->rendered_object->width));
			y_diff = y_pos - (y +
				(d->halignment/2)*(drawable_object->rendered_object->height));

			alloc->y -= y_diff;
			if (d->halignment == 2) {
				if (x_diff > 0) {
					*offset_x += x_diff;
				} else {
					alloc->x -= x_diff;
				}
			} else {
				if (x_diff < 0) {
					*offset_x += x_diff;
				} else {
					alloc->x -= x_diff;
				}
			}
		}
	}
}

KibaDock* kiba_visual_feature_get_dock(KibaVisualFeature *feature)
{
	g_return_if_fail(KIBA_IS_VISUAL_FEATURE(feature));
	
	KibaVisualFeaturePrivate *priv = KIBA_VISUAL_FEATURE_GET_PRIVATE (feature);
	if (priv->win) {
		return KIBA_WIN_DOCK(priv->win);
	} else {
		return NULL;
	}
}

KibaWin* kiba_visual_feature_get_win(KibaVisualFeature *feature)
{
	g_return_if_fail(KIBA_IS_VISUAL_FEATURE(feature));
	return KIBA_VISUAL_FEATURE_GET_PRIVATE (feature)->win;
}

void kiba_visual_feature_set_win(KibaVisualFeature *feature, KibaWin* win)
{
	g_return_if_fail(KIBA_IS_VISUAL_FEATURE(feature));
	g_return_if_fail(KIBA_IS_WIN(win));
	
	KibaVisualFeaturePrivate *priv = KIBA_VISUAL_FEATURE_GET_PRIVATE (feature);
	
	priv->win = win;
	kiba_win_add_visual_feature(priv->win, feature);
	
	g_signal_connect(G_OBJECT(priv->win), "add",
		G_CALLBACK(__dock_added_cb), feature);
	g_signal_connect(G_OBJECT(priv->win), "remove",
		G_CALLBACK(__dock_removed_cb), feature);
	if (KIBA_WIN_DOCK(win) != NULL) {
		__dock_added_cb(GTK_CONTAINER(win),
			GTK_WIDGET(KIBA_WIN_DOCK(win)), feature);
	}
}

void kiba_visual_feature_object_added(KibaVisualFeature *feature,
		KibaObject *object)
{
	g_return_if_fail(KIBA_IS_VISUAL_FEATURE(feature));
	g_return_if_fail(object != NULL);
	
	feature->object_added(feature, object);
}

void kiba_visual_feature_object_removed(KibaVisualFeature *feature,
		KibaObject *object)
{
	g_return_if_fail(KIBA_IS_VISUAL_FEATURE(feature));
	g_return_if_fail(object != NULL);
	
	feature->object_removed(feature, object);
}

void kiba_visual_feature_draw(KibaVisualFeature *feature,
		KibaVisualFeatureObject *object,
		cairo_t *cr)
{
	g_return_if_fail(KIBA_IS_VISUAL_FEATURE(feature));
	g_return_if_fail(object != NULL);
	
	feature->draw(feature, object, cr);
}

void kiba_visual_feature_queue_draw(KibaVisualFeature *feature,
		KibaVisualFeatureObject *object)
{
	g_return_if_fail(KIBA_IS_VISUAL_FEATURE(feature));
	g_return_if_fail(object != NULL);
	
	KibaVisualFeaturePrivate *priv = KIBA_VISUAL_FEATURE_GET_PRIVATE (feature);
	__invalidate_kiba_alloc(priv->win, object->abs_allocation);
}

void kiba_visual_feature_queue_resize(KibaVisualFeature *feature,
		KibaVisualFeatureObject *object)
{
	g_return_if_fail(KIBA_IS_VISUAL_FEATURE(feature));
	g_return_if_fail(object != NULL);
	
	object->queue_resize = TRUE;
}

void kiba_visual_feature_do_resize(KibaVisualFeature *feature,
		KibaVisualFeatureObject *object)
{
	if (!object->queue_resize) {
		return;
	}
	object->queue_resize = FALSE;
	
	KibaVisualFeaturePrivate *priv = KIBA_VISUAL_FEATURE_GET_PRIVATE (feature);
	KibaAllocation old = object->abs_allocation;
	KibaAllocation old_rel = object->allocation;
	gboolean size_changed;
	
	__get_allocation(feature, object);
	KibaAllocation new = object->abs_allocation;
	size_changed = old.width != new.width || old.height != new.height;
	if (old.x != new.x || old.y != new.y || size_changed) {
		__invalidate_kiba_alloc(priv->win, old);
		__invalidate_kiba_alloc(priv->win, new);
	}
	
	//if (priv->type == KIBA_VISUAL_FEATURE_ONTOP_OF_DOCK) {
	//	return;
	//}
	
	if (size_changed || old_rel.x != object->allocation.x || old_rel.y != object->allocation.y) {
		switch(priv->type) {
		case KIBA_VISUAL_FEATURE_MORPH:
		case KIBA_VISUAL_FEATURE_ONTOP:
			kiba_drawable_object_queue_render_object(KIBA_DRAWABLE_OBJECT(object->object));
			break;
		case KIBA_VISUAL_FEATURE_GIMMICK:
			break;
		default:
			break;
		}
	}
}

void kiba_visual_feature_do_resize_all(KibaVisualFeature *feature)
{
	g_return_if_fail(KIBA_IS_VISUAL_FEATURE(feature));
	
	GList *l;
	for (l = feature->objects; l != NULL; l = l->next) {
		kiba_visual_feature_do_resize(feature,
			(KibaVisualFeatureObject*) l->data);
	}
}

void kiba_visual_feature_queue_draw_all(KibaVisualFeature *feature)
{
	g_return_if_fail(KIBA_IS_VISUAL_FEATURE(feature));
	
	GList *l;
	for (l = feature->objects; l != NULL; l = l->next) {
		kiba_visual_feature_queue_draw(feature,
			(KibaVisualFeatureObject*) l->data);
	}
}

void kiba_visual_feature_queue_resize_all(KibaVisualFeature *feature)
{
	g_return_if_fail(KIBA_IS_VISUAL_FEATURE(feature));
	
	GList *l;
	for (l = feature->objects; l != NULL; l = l->next) {
		kiba_visual_feature_queue_resize(feature,
			(KibaVisualFeatureObject*) l->data);
	}
}

KibaVisualFeatureType kiba_visual_feature_type (KibaVisualFeature *feature)
{
	g_return_if_fail(KIBA_IS_VISUAL_FEATURE(feature));
	return KIBA_VISUAL_FEATURE_GET_PRIVATE (feature)->type;
}

gboolean kiba_visual_feature_get_front (KibaVisualFeature *feature)
{
	g_return_if_fail(KIBA_IS_VISUAL_FEATURE(feature));
	return KIBA_VISUAL_FEATURE_GET_PRIVATE (feature)->front;
}

const gchar* kiba_visual_feature_get_name (KibaVisualFeature *feature)
{
	g_return_if_fail(KIBA_IS_VISUAL_FEATURE(feature));
	return KIBA_VISUAL_FEATURE_GET_PRIVATE (feature)->name;
}

KibaVisualFeatureObject* kiba_visual_feature_add_object(KibaVisualFeature *feature, KibaObject *object)
{
	g_return_if_fail(KIBA_IS_VISUAL_FEATURE(feature));
	
	KibaVisualFeaturePrivate *priv = KIBA_VISUAL_FEATURE_GET_PRIVATE (feature);
	KibaVisualFeatureObject *feature_object;
	
	feature_object = g_new0 (KibaVisualFeatureObject, 1);
	feature->objects = g_list_prepend(feature->objects, feature_object);
	
	feature_object->feature = feature;
	feature_object->object = object;
	feature_object->queue_resize = TRUE;
	feature_object->allocation = __get_allocation(feature, feature_object);
	
	g_signal_connect(G_OBJECT(object), "unrealize",
		G_CALLBACK(__destroy_cb), feature_object);
	
	kiba_visual_feature_queue_draw(feature, feature_object);
	
	kiba_object_enable_feature(object, feature_object);
	
	// the win size may changed
	gtk_widget_queue_resize_no_redraw(GTK_WIDGET(priv->win));
	gtk_widget_queue_resize_no_redraw(GTK_WIDGET(KIBA_WIN_DOCK(priv->win)));
	
	switch(priv->type) {
	case KIBA_VISUAL_FEATURE_MORPH:
	case KIBA_VISUAL_FEATURE_ONTOP:
		kiba_drawable_object_queue_render_object(KIBA_DRAWABLE_OBJECT(object));
		break;
	case KIBA_VISUAL_FEATURE_GIMMICK:
		break;
	default:
		break;
	}
	
	return feature_object;
}

void kiba_visual_feature_remove_object(KibaVisualFeature *feature, KibaObject *object)
{
	g_return_if_fail(KIBA_IS_VISUAL_FEATURE(feature));
	
	KibaVisualFeaturePrivate *priv = KIBA_VISUAL_FEATURE_GET_PRIVATE (feature);
	KibaVisualFeatureObject *feature_object = NULL;
	GList *l;
	
	for (l = feature->objects; l != NULL; l = l->next) {
		if (((KibaVisualFeatureObject*) l->data)->object == object) {
			feature_object = l->data;
			break;
		}
	}
	
	if(feature_object == NULL)
		return;
	
	g_signal_handlers_disconnect_by_func (G_OBJECT(object),
		G_CALLBACK(__destroy_cb), feature_object);
	
	kiba_object_disable_feature(object, feature_object);
	kiba_visual_feature_queue_draw(feature, feature_object);
	feature->objects = g_list_remove(feature->objects, feature_object);
	g_free(feature_object);
	
	// the win size may changed
	gtk_widget_queue_resize_no_redraw(GTK_WIDGET(KIBA_WIN_DOCK(priv->win)));
	gtk_widget_queue_resize_no_redraw(GTK_WIDGET(priv->win));
	
	switch(priv->type) {
	case KIBA_VISUAL_FEATURE_MORPH:
	case KIBA_VISUAL_FEATURE_ONTOP:
		kiba_drawable_object_queue_render_object(KIBA_DRAWABLE_OBJECT(object));
		break;
	case KIBA_VISUAL_FEATURE_GIMMICK:
		break;
	default:
		break;
	}
}

/*** Class implementation */

static void kiba_visual_feature_object_added_impl(KibaVisualFeature *feature,
		KibaObject *object)
{
	// this method must be implemented by the subclass
	g_warning ("KibaVisualFeatureClass::object_added not implemented for `%s'",
		kiba_visual_feature_get_name (feature));
}

static void kiba_visual_feature_object_removed_impl(KibaVisualFeature *feature,
		KibaObject *object)
{
	// this method must be implemented by the subclass
	g_warning ("KibaVisualFeatureClass::object_removed not implemented for `%s'",
		kiba_visual_feature_get_name (feature));
}

static void kiba_visual_feature_draw_impl(KibaVisualFeature *feature,
		KibaVisualFeatureObject *object,
		cairo_t *cr)
{
	// this method must be implemented by the subclass
	g_warning ("KibaVisualFeatureClass::draw not implemented for `%s'",
		kiba_visual_feature_get_name (feature));
}

static KibaAllocation kiba_visual_feature_get_allocation_impl(KibaVisualFeature *feature,
		KibaVisualFeatureObject *object)
{
	// this method must be implemented by the subclass
	g_warning ("KibaVisualFeatureClass::get_allocation not implemented for `%s'",
		kiba_visual_feature_get_name (feature));
	return (KibaAllocation) {0.0f, 0.0f, 0.0f, 0.0f};
}

static void kiba_visual_feature_get_max_offset(KibaVisualFeature *feature,
		KibaVisualFeatureObject *object,
		float *left, float *right,
		float *top, float *bottom)
{
	*left = 0.0f;
	*right = 0.0f;
	*top = 0.0f;
	*bottom = 0.0f;
}

static void kiba_visual_feature_dispose (GObject *obj)
{
	KibaVisualFeature *feature = KIBA_VISUAL_FEATURE(obj);
	KibaVisualFeaturePrivate *priv = KIBA_VISUAL_FEATURE_GET_PRIVATE (obj);
	KibaDock *dock = KIBA_WIN_DOCK(priv->win);
	GList *l, *children;
	
	g_signal_handlers_disconnect_by_func(G_OBJECT(priv->win),
		G_CALLBACK(__dock_added_cb), feature);
	g_signal_handlers_disconnect_by_func(G_OBJECT(priv->win),
		G_CALLBACK(__dock_removed_cb), feature);
	g_signal_handlers_disconnect_by_func(G_OBJECT(dock),
		G_CALLBACK(__object_added), feature);
	g_signal_handlers_disconnect_by_func(G_OBJECT(dock),
		G_CALLBACK(__object_removed), feature);
	
	children = gtk_container_get_children(GTK_CONTAINER(dock));
	for (l = children; l != NULL; l = children) {
		GtkWidget* object = l->data;
		children = g_list_remove(children, object);
		__object_removed(GTK_CONTAINER(dock), GTK_WIDGET(object), feature);
	}
	g_list_free(children);
}

/**
 * g_object_set_property implementation.
 **/
static void kiba_visual_feature_set_property (GObject *object,
                                      guint property_id,
                                      const GValue *value,
                                      GParamSpec *pspec) {
	KibaVisualFeature *self = KIBA_VISUAL_FEATURE(object);
	KibaVisualFeaturePrivate *priv = KIBA_VISUAL_FEATURE_GET_PRIVATE (self);

	switch (property_id) {
		case _PROP_NAME:
			priv->name = g_value_dup_string (value);
			break;
		case _PROP_TYPE:
			priv->type = g_value_get_int (value);
			break;
		case _PROP_FRONT:
			priv->front = g_value_get_boolean (value);
			if (priv->win != NULL && GTK_BIN(priv->win)->child != NULL)
				__dock_added_cb(GTK_CONTAINER(priv->win), GTK_BIN(priv->win)->child, self);
			break;
		default:
			G_OBJECT_WARN_INVALID_PROPERTY_ID(object,property_id,pspec);
			break;
	}
}

/**
 * g_object_get_property implementation.
 **/
static void kiba_visual_feature_get_property (GObject *object,
                                      guint property_id,
                                      GValue *value,
                                      GParamSpec *pspec) {
	KibaVisualFeature *self = KIBA_VISUAL_FEATURE(object);
	KibaVisualFeaturePrivate *priv = KIBA_VISUAL_FEATURE_GET_PRIVATE (self);

	switch (property_id) {
		case _PROP_NAME:
			g_value_set_string (value, priv->name);
			break;
		case _PROP_TYPE:
			g_value_set_int (value, priv->type);
			break;
		case _PROP_FRONT:
			g_value_set_boolean (value, priv->front);
			break;
		default:
			G_OBJECT_WARN_INVALID_PROPERTY_ID(object,property_id,pspec);
			break;
	}
}

static void kiba_visual_feature_class_init (KibaVisualFeatureClass *class)
{
	GObjectClass *gobject_class = G_OBJECT_CLASS (class);
	
	// GObject overrides
	gobject_class->set_property = kiba_visual_feature_set_property;
	gobject_class->get_property = kiba_visual_feature_get_property;
	gobject_class->dispose = kiba_visual_feature_dispose;
	
	// properties
	g_object_class_install_property
		(gobject_class,
		 _PROP_NAME,
		 g_param_spec_string ("name",
		 "Name",
		 "The feature name",
		 "",
		 G_PARAM_CONSTRUCT_ONLY|G_PARAM_READWRITE));
	g_object_class_install_property
		(gobject_class,
		 _PROP_TYPE,
		 g_param_spec_int ("type",
		 "Type",
		 "The feature type",
		 KIBA_VISUAL_FEATURE_FIRST,
		 KIBA_VISUAL_FEATURE_LAST-1,
		 KIBA_VISUAL_FEATURE_GIMMICK,
		 G_PARAM_CONSTRUCT_ONLY|G_PARAM_READWRITE));
	g_object_class_install_property
		(gobject_class,
		 _PROP_FRONT,
		 g_param_spec_boolean ("front",
		 "Front",
		 "Wether the feature is in front of objects",
		 TRUE,
		 G_PARAM_CONSTRUCT_ONLY|G_PARAM_READWRITE));
	
	// private class
	g_type_class_add_private (gobject_class, sizeof (KibaVisualFeaturePrivate));
}

static void kiba_visual_feature_init (KibaVisualFeature *feature)
{
	KibaVisualFeaturePrivate *priv = KIBA_VISUAL_FEATURE_GET_PRIVATE (feature);
	
	feature->objects = NULL;
	
	priv->type = KIBA_VISUAL_FEATURE_FIRST;
	
	// KibaVisualFeature overrides
	feature->get_allocation = kiba_visual_feature_get_allocation_impl;
	feature->draw = kiba_visual_feature_draw_impl;
	feature->object_added = kiba_visual_feature_object_added_impl;
	feature->object_removed = kiba_visual_feature_object_removed_impl;
	feature->get_max_offset = kiba_visual_feature_get_max_offset;
	feature->win_set = NULL;
}

KibaVisualFeature* kiba_visual_feature_new(
	KibaWin *win,
	const char *name,
	KibaVisualFeatureType type,
	gboolean front)
{
	return g_object_new (KIBA_TYPE_VISUAL_FEATURE,
		    "win",   win,
		    "name",  name,
		    "type",  type,
		    "front", front,
		    NULL);
}
