/*
 *  Copyright (C) 2006-2009 Daniel Bessler <danielb@kiba-dock.org>
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110-1301  USA.
 *
 *  Author : Daniel Bessler <danielb@kiba-dock.org>
*/

#include "../config.h"
#include <unistd.h>
#include <stdio.h>
#include <stdlib.h>
#include <gtk/gtk.h>
#include <gdk/gdk.h>
#include <gdk/gdkx.h>
#include <gdk-pixbuf/gdk-pixbuf.h>
#include <gdk-pixbuf-xlib/gdk-pixbuf-xlib.h>
#include <X11/Xlib.h>
#include <X11/Xatom.h>
#if GTK_CHECK_VERSION(2,9,0) == FALSE
  #include <X11/extensions/shape.h>
#endif

#include "kiba.h"
#include "kiba-win.h"
#include "kiba-object.h"
#include "kiba-screen.h"
#include "kiba-debug.h"
#include "kiba-x11.h"
#include "kiba-separator.h"
#include "kiba-macros.h"
#include "kiba-cairo-utils.h"
#include "greyed.h"
#ifdef USE_GLITZ
  #include "glitz-helper.h"
#endif
#ifndef USE_GLITZ
  #define USE_OFFSCREEN_PIXMAP
#endif

#include "kiba-private.h"

// TRUE for benchmarking
#define BENCHMARK_ENABLE FALSE
// TRUE to see only the rerendered parts of the window
#define CLIPPING_CHECK FALSE

#define KIBA_WIN_GET_PRIVATE(obj) \
		(G_TYPE_INSTANCE_GET_PRIVATE ((obj), KIBA_TYPE_WIN, KibaWinPrivate))

#define WIN_DRAG_AREA 10000
// area around the screen middle where the dock snaps to the center
#define DOCK_SNAP_AREA 15
// alpha while dragging the dock
#define DOCK_DRAG_ALPHA 0.4
// max frames per second for rendering
#define MAX_FPS 50
// timeout rates
#define REDRAW_RATE 1000/MAX_FPS
#define CHECK_HIDE_RATE 50

#define ATOM(a) XInternAtom(display, #a, False)

// preference keys
#define PREFERENCES_BELOW_KEY "keep_window_below_others"
#define PREFERENCES_ABOVE_KEY "keep_window_above_others"
#define PREFERENCES_HIDE_DELAY_KEY "auto_hide_delay"
#define PREFERENCES_HIDE_SPEED_KEY "auto_hide_speed"
#define PREFERENCES_HIDE_AREA_KEY "auto_hide_area"
#define PREFERENCES_EDGE_OFFSET_KEY "dock_edge_offset"

G_DEFINE_TYPE (KibaWin, kiba_win, KIBA_TYPE_TRANSPARENT_WINDOW)

typedef struct _KibaWinPrivate KibaWinPrivate;

/**
 * KibaWinDragData:
 * @receive_func: the method to handle the drag data on receive events.
 * @dest_func: the method to check if receive_func is able to handle the drag.
 * @user_data: the user data pointer for the methods.
 *
 * Type to handle a drag and drop receice event.
 */
typedef struct {
	KibaDragReceiveFunc receive_func;
	KibaDragDestFunc dest_func;
	gpointer user_data;
}KibaWinDragData;

/*
 * Type used to represent a drag and drop receive method.
 */
typedef struct {
	KibaWinDragData *data;
	GList *actions;
	GdkDragContext *context;
	gchar *drag_data;
	gchar *action;
	gchar *icon;
}KibaDragAction;


struct _KibaWinPrivate {
	/* damaged area */
	GdkRegion *invalid_region;
	/* id for the framerate limiter timeout */
	gint redraw_id;

	gint move_dock_id; /* id for the dock drag timeout */
	gint check_hide_id; /* id for the timeout checking the mouse position for hiding */
	gint hide_id; /* id for the timeout hiding the dock */
	gint raise_id;

	GdkRectangle mouse_anchor;  /* saves last mouse position for dragging */
	GList *drag_reveive_data; /* list of custom dnd receive functions (KibaWinDragData) */
	
	GList *features;

	gboolean composited; /* true if composite manager running */
	gint depth; /* depth of the visual for this window */
	
	GtkAllocation choosen_allocation;
	GtkAllocation last_allocation;
	
#ifdef USE_GLITZ
	glitz_drawable_format_t *glitz_draw_format;
	glitz_drawable_t *glitz_drawable;
	glitz_format_t *glitz_format;
#endif
#ifdef USE_OFFSCREEN_PIXMAP
	GdkPixmap *pixmap;
#endif
};

// signal ids
enum {
	SIG_KIBA_HIDDEN = 0,
	SIG_KIBA_UNHIDDEN,
	SIG_KIBA_WIN_LAST
};

static guint kiba_win_signals[SIG_KIBA_WIN_LAST] = { 0 };

static gint __redraw (gpointer data);

#if BENCHMARK_ENABLE == TRUE
static gint redraw_count = 0;
#endif

// preferences info
static PrefMember win_options[] = {
	{ PREFERENCES_BELOW_KEY, "FALSE", KIBA_BOOLEAN },
	{ PREFERENCES_ABOVE_KEY, "FALSE", KIBA_BOOLEAN },
	{ PREFERENCES_HIDE_DELAY_KEY, "5", KIBA_INT },
	{ PREFERENCES_HIDE_SPEED_KEY, "5", KIBA_INT },
	{ PREFERENCES_EDGE_OFFSET_KEY, "0", KIBA_INT },
	{ PREFERENCES_HIDE_AREA_KEY, "100", KIBA_INT }
};

/*** Helper Functions ***/

#if BENCHMARK_ENABLE == TRUE
static gint __benchmark_cb (gpointer data) {
	fprintf (stderr, "Benchmark: %i FPS\n", redraw_count);
	redraw_count = 0;
	return TRUE;
}
#endif

/*
 * Returns the closest edge position related to the coordinates x/y.
 */
static int
__find_closest_dock_position (GtkWidget *dock,
		gint x, gint y)
{
	gint i, dx, dy, min_d, d, position;

	position = -1;

	min_d = WIN_DRAG_AREA;
	for (i = 0; i < G_N_ELEMENTS (kiba_layout_data); i++) {
		dx = x - KIBA_SCREEN_DEFAULT->geometry.width*kiba_layout_data[i].anchor.x/4;
		dy = y - KIBA_SCREEN_DEFAULT->geometry.height*kiba_layout_data[i].anchor.y/4;

		d = dx * dx + dy * dy;
		if (d < min_d) {
			min_d    = d;
			position = i;
		}
	}
	
	return position;
}

/*
 * makes a list of free spaces. first starts with
 * one element and complete screen width.
 * then breaking it into parts, where docks are placed.
 */
static GList*
__get_free_spaces(KibaDock *dock,
		KibaPosition position,
		GtkOrientation orient,
		GtkAllocation alloc)
{
	GtkWidget *win = GTK_WIDGET(dock)->parent;
	GList *l, *m, *docks;
	gint *val = (gint*) g_new0(gint, 2);
	GList *ret = NULL;

	if (orient == GTK_ORIENTATION_VERTICAL) {
		val[0] = KIBA_SCREEN_DEFAULT->geometry.y - dock->diff_l;
		val[1] = KIBA_SCREEN_DEFAULT->geometry.y +
				KIBA_SCREEN_DEFAULT->geometry.height + dock->diff_r;
	} else {
		val[0] = KIBA_SCREEN_DEFAULT->geometry.x - dock->diff_l;
		val[1] = KIBA_SCREEN_DEFAULT->geometry.x +
				KIBA_SCREEN_DEFAULT->geometry.width + dock->diff_r;
	}
	ret = g_list_prepend(NULL, val);

	docks = kiba_get_docks(kiba_get());
	for (l = docks; l != NULL; l = docks) {
		KibaDock *buf = l->data;
		GtkWidget *wBuf = GTK_WIDGET(buf)->parent;
		docks = g_list_remove(docks, l->data);
		KibaPosition bufPosition = kiba_dock_get_position(buf);
		GtkOrientation bufOrient =
			kiba_layout_data[bufPosition].orientation;

		if (dock == buf)
			continue;

		gint space[2];
		if (bufPosition == position) {
			if (orient == GTK_ORIENTATION_VERTICAL) {
				space[0] = wBuf->allocation.y;
				space[1] = wBuf->allocation.y + wBuf->allocation.height;
			} else {
				space[0] = wBuf->allocation.x;
				space[1] = wBuf->allocation.x + wBuf->allocation.width;
			}
		} else if (bufOrient != orient) {
			gboolean collides = FALSE;
			switch(position) {
			case KIBA_POSITION_TOP_EDGE:
				if (wBuf->allocation.y < win->allocation.y + win->allocation.height) {
					if (bufPosition == KIBA_POSITION_LEFT_EDGE) {
						space[0] = KIBA_SCREEN_DEFAULT->geometry.x - dock->diff_l;
						space[1] = space[0] + wBuf->allocation.width;
					} else {
						space[1] = KIBA_SCREEN_DEFAULT->geometry.x +
								KIBA_SCREEN_DEFAULT->geometry.width + dock->diff_r;
						space[0] = space[1] - wBuf->allocation.width;
					}
					collides = TRUE;
				}
				break;
			case KIBA_POSITION_BOTTOM_EDGE:
				if (wBuf->allocation.y + wBuf->allocation.height > win->allocation.y) {
					if (bufPosition == KIBA_POSITION_LEFT_EDGE) {
						space[0] = KIBA_SCREEN_DEFAULT->geometry.x - dock->diff_l;
						space[1] = space[0] + wBuf->allocation.width;
					} else {
						space[1] = KIBA_SCREEN_DEFAULT->geometry.x +
							KIBA_SCREEN_DEFAULT->geometry.width + dock->diff_r;
						space[0] = space[1] - wBuf->allocation.width;
					}
					collides = TRUE;
				}
				break;
			case KIBA_POSITION_LEFT_EDGE:
				if (wBuf->allocation.x < win->allocation.x + win->allocation.width) {
					if (bufPosition == KIBA_POSITION_TOP_EDGE) {
						space[0] = KIBA_SCREEN_DEFAULT->geometry.y - dock->diff_l;
						space[1] = space[0] + wBuf->allocation.height;
					} else {
						space[1] = KIBA_SCREEN_DEFAULT->geometry.y +
								KIBA_SCREEN_DEFAULT->geometry.height + dock->diff_r;
						space[0] = space[1] - wBuf->allocation.height;
					}
					collides = TRUE;
				}
				break;
			case KIBA_POSITION_RIGHT_EDGE:
				if (wBuf->allocation.x + wBuf->allocation.width > win->allocation.x) {
					if (bufPosition == KIBA_POSITION_TOP_EDGE) {
						space[0] = KIBA_SCREEN_DEFAULT->geometry.y - dock->diff_l;
						space[1] = space[0] + wBuf->allocation.height;
					} else {
						space[1] = KIBA_SCREEN_DEFAULT->geometry.y +
								KIBA_SCREEN_DEFAULT->geometry.height + dock->diff_r;
						space[0] = space[1] - wBuf->allocation.height;
					}
					collides = TRUE;
				}
				break;
			default:
				g_assert_not_reached();
			}
			if (!collides) continue;
		} else {
			continue;
		}

		GList *retCopy = g_list_copy(ret);
		for (m = retCopy; m != NULL; m = retCopy) {
			gint *y = m->data;
			gint *buf;

			retCopy = g_list_remove(retCopy, m->data);
			if (y[0] > space[0]) {
				if (y[0] > space[1])
					continue; // desired space is left of
				ret = g_list_remove(ret, y);
				if (space[1] >= y[1]) {
					// space consumes whole free space
					free(y);
					continue;
				}

				buf = (gint*) g_new0(gint, 2);
				buf[0] = space[0];
				buf[1] = y[1];
				if (buf[1]-buf[0] > dock->d.object_size)
					ret = g_list_prepend(ret, buf);
				else
					free(buf);
				free(y);
			} else if (y[1] > space[0]) {
				ret = g_list_remove(ret, y);
				buf = (gint*) g_new0(gint, 2);
				buf[0] = y[0];
				buf[1] = space[0];
				if (buf[1]-buf[0] > dock->d.object_size)
					ret = g_list_prepend(ret, buf);
				else
					free(buf);
				if (y[1] > space[1]) {
					// space cuts free space in two pieces
					buf = (gint*) g_new0(gint, 2);
					buf[0] = space[1];
					buf[1] = y[1];
					if (buf[1]-buf[0] > dock->d.object_size)
						ret = g_list_prepend(ret, buf);
					else
						free(buf);
				}
				free(y);
			}
		}
	}
	g_list_free(docks);

	return ret;
}

/** find the nearest space in a list of spaces */
static void
__get_nearest_space(gint *wanted,
		GList *freeList,
		gint *orient_ret,
		gint *nearestSpace_ret)
{
	GList *l;
	gint *nearest = NULL;
	gint space;
	// iterate free spaces
	for (l = freeList; l != NULL; l = l->next) {
		gint *near = l->data;
		gint mOrient = (near[0]+near[1] > wanted[0]+wanted[1] ? 0 : 1);
		gint mSpace;
		// get space covered by wanted and free space
		if (mOrient == 0)
			mSpace = wanted[1]-near[0];
		else
			mSpace = near[1]-wanted[0];
		// if the space is maximal, we use this space
		if (!nearest || mSpace>space) {
			nearest = near;
			space = mSpace;
			*orient_ret = mOrient;
		}
	}
	if (nearest) {
		nearestSpace_ret[0] = nearest[0];
		nearestSpace_ret[1] = nearest[1];
	} else {
		// FIXME: if nearest==NULL, there is no space left.
		// we could ask the user to use another side of the screen or
		// to delete the dock.
		// or the other docks could be shrinked.
		nearestSpace_ret[0] = wanted[0];
		nearestSpace_ret[1] = wanted[1];
	}
}

/** fit a wanted space into another space */
static void
__fit_to_space(gint *wanted,
		gint *space,
		gint *ret,
		gint orient)
{
	gint wSize = wanted[1]-wanted[0];
	if (orient == 0) {
		// right / below of
		ret[0] = space[0] < wanted[0] ? wanted[0] : space[0];
		ret[1] = MIN(space[1], ret[0]+wSize);
	} else {
		// left / above of
		ret[1] = space[1] > wanted[1] ? wanted[1] : space[1];
		ret[0] = MAX(space[0], ret[1]-wSize);
	}
}

static void
__choose_allocation(KibaWin *win,
		GtkRequisition *req)
{
	KibaWinPrivate *priv = KIBA_WIN_GET_PRIVATE (win);
	KibaDock *dock = KIBA_WIN_DOCK(win);
	const KibaLayoutData *d;
	gint wantedSpace[2], nearestSpace[2], choosenSpace[2];
	GList *l;
	
	// no dock set yet
	if (!KIBA_IS_DOCK(dock)) return;
	gint edge_offset = (win->offset < DOCK_SNAP_AREA &&
			win->offset > -DOCK_SNAP_AREA ?
			0 : win->offset) - dock->diff_l;
	d = &kiba_layout_data[dock->d.position];
	
	if (d->orientation == GTK_ORIENTATION_VERTICAL) {
		priv->choosen_allocation.height = MIN(dock->d.max_size, req->height);
		priv->choosen_allocation.width = req->width;

		priv->choosen_allocation.x = KIBA_SCREEN_DEFAULT->geometry.x +
			d->halignment/2 * (KIBA_SCREEN_DEFAULT->geometry.width - 
			priv->choosen_allocation.width);
		priv->choosen_allocation.y = KIBA_SCREEN_DEFAULT->geometry.y +
			0.5f*(KIBA_SCREEN_DEFAULT->geometry.height -
			priv->choosen_allocation.height) +
			dock->diff_r + edge_offset;
		wantedSpace[0] = priv->choosen_allocation.y;
		wantedSpace[1] = priv->choosen_allocation.y + priv->choosen_allocation.height;
	} else {
		priv->choosen_allocation.width = MIN(dock->d.max_size, req->width);
		priv->choosen_allocation.height = req->height;
		
		priv->choosen_allocation.y = KIBA_SCREEN_DEFAULT->geometry.y +
			d->valignment/2 * (KIBA_SCREEN_DEFAULT->geometry.height -
			priv->choosen_allocation.height);
		priv->choosen_allocation.x = KIBA_SCREEN_DEFAULT->geometry.x +
			0.5f*(KIBA_SCREEN_DEFAULT->geometry.width -
			priv->choosen_allocation.width) +
			dock->diff_r + (dock->diff_l - dock->diff_r)/2 + edge_offset;
		wantedSpace[0] = priv->choosen_allocation.x;
		wantedSpace[1] = priv->choosen_allocation.x + priv->choosen_allocation.width;
	}

	GList* freeSpace = __get_free_spaces(dock,
			kiba_dock_get_position(dock),
			d->orientation,
			priv->choosen_allocation);
	gint orient = 0;
	__get_nearest_space(wantedSpace, freeSpace, &orient, nearestSpace);
	__fit_to_space(wantedSpace, nearestSpace, choosenSpace, orient);
	for (l = freeSpace; l != NULL; l = freeSpace) {
		free((gfloat*) l->data);
		freeSpace = g_list_remove(freeSpace, l->data);
	}
	g_list_free(freeSpace);
	
	gint dx = (wantedSpace[0] + wantedSpace[1]/2) -
		    (choosenSpace[0] + choosenSpace[1]/2);
	if (dx == 0) {
		win->dockHeading = KIBA_DOCK_HEADING_NONE;
	} else if (dx < 0) {
		win->dockHeading = KIBA_DOCK_HEADING_LEFT;
	} else {
		win->dockHeading = KIBA_DOCK_HEADING_RIGHT;
	}
	win->dockSpace[0] =  choosenSpace[0];
	win->dockSpace[1] =  choosenSpace[1];
	
	if (d->orientation == GTK_ORIENTATION_VERTICAL) {
		priv->choosen_allocation.y = choosenSpace[0];
		priv->choosen_allocation.height = choosenSpace[1] - choosenSpace[0];
	} else {
		priv->choosen_allocation.x = choosenSpace[0];
		priv->choosen_allocation.width = choosenSpace[1] - choosenSpace[0];
	}
	
	gtk_widget_set_usize(GTK_WIDGET(win),
		priv->choosen_allocation.width, priv->choosen_allocation.height);
}

/*
 * allocates object size
 */
static void
__resize_children (KibaWin *win)
{
	Kiba *kiba = kiba_get();
	GtkWidget *widget = GTK_WIDGET(win);
	GtkWidget *dock = KIBA_WIN_CHILD(widget);
	KibaWinPrivate *priv = KIBA_WIN_GET_PRIVATE (win);
	const KibaLayoutData *d;
	GtkRequisition req;
	GtkAllocation allocation;

	d = &kiba_layout_data[KIBA_DOCK(dock)->d.position];
	allocation = dock->allocation;
	
	// get the size request from the dock
	gtk_widget_size_request (dock, &req);
	if (req.width || req.height) {
		// non-empty window
		allocation.width = req.width;
		allocation.height = req.height;
	}

	if (d->orientation == GTK_ORIENTATION_VERTICAL) {
		allocation.height = MIN(allocation.height, KIBA_DOCK(dock)->d.max_size);
		allocation.y = 0;
	} else {
		allocation.width = MIN(allocation.width, KIBA_DOCK(dock)->d.max_size);
		allocation.x = 0;
	}
	
	allocation.width = MIN(priv->choosen_allocation.width, MAX(allocation.width, 0));
	allocation.height = MIN(priv->choosen_allocation.height, MAX(allocation.height, 0));
	gtk_widget_size_allocate (dock, &allocation);
}

/**
 * Moves dock on button presses.
 **/
static gint
__move_dock (gpointer data)
{
	KibaWin *win = data;
	GtkWidget *dock = KIBA_WIN_CHILD(win);
	KibaWinPrivate *priv = KIBA_WIN_GET_PRIVATE (win);
	KibaPosition position;
	const KibaLayoutData *d;
	GdkScreen *new_screen = NULL;
	gint x = 0, y = 0, monitor;
	GdkRectangle monitor_size;
	
	// get pointer relative to active screen on display
	gdk_display_get_pointer (
		KIBA_SCREEN_DISPLAY,
		&new_screen,
		&x, &y, NULL);
	if (KIBA_GDK_SCREEN != new_screen) {
		// screen changed
#if 0
		// FIXME: changing the screen needs some more work,
		//        the window just disappeares.
		kiba_screen_set_gdk_screen(KIBA_SCREEN_DEFAULT, new_screen);
		gtk_window_set_screen(GTK_WINDOW(win), new_screen);
		
		// screen should have changed
		g_assert(KIBA_GDK_SCREEN == new_screen);
#else
		return TRUE;
#endif
	}
	
	monitor = gdk_screen_get_monitor_at_point (KIBA_GDK_SCREEN, x, y);
	gdk_screen_get_monitor_geometry (KIBA_GDK_SCREEN,
		monitor, &monitor_size);
	
	// get pointer relative to root window
	gdk_window_get_pointer (
		gdk_screen_get_root_window(KIBA_GDK_SCREEN),
		&x, &y, NULL);
	x -= monitor_size.x;
	y -= monitor_size.y;
	
	position = __find_closest_dock_position (dock, x, y);
	d = &kiba_layout_data[position==-1 ? KIBA_DOCK(dock)->d.position : position];
	gboolean monitor_changed = kiba_screen_get_monitor(KIBA_SCREEN_DEFAULT) != monitor;
	
	if (position != -1 && (position != KIBA_DOCK(dock)->d.position || monitor_changed)) {
		win->offset = 0;
		kiba_screen_set_monitor(KIBA_SCREEN_DEFAULT, monitor);
		kiba_dock_set_position (KIBA_DOCK(dock), position);
	} else if (!monitor_changed) {
		if (d->orientation == GTK_ORIENTATION_VERTICAL)
			win->offset += y - priv->mouse_anchor.y;
		else
			win->offset += x - priv->mouse_anchor.x;
		gtk_widget_queue_resize_no_redraw(GTK_WIDGET(win));
	}

	priv->mouse_anchor.x = x;
	priv->mouse_anchor.y = y;

	return TRUE;
}

static gint
__hide_dock (gpointer data)
{
	KibaWin *win = KIBA_WIN(data);
	GtkWidget *widget = GTK_WIDGET(win);
	GtkWidget *dock = KIBA_WIN_CHILD(win);
	KibaWinPrivate *priv = KIBA_WIN_GET_PRIVATE (win);
	GtkAllocation allocation;
	const KibaLayoutData *d;
	gboolean hidden = FALSE;

	if (win->autohide.state != KIBA_HIDING) {
		priv->hide_id = 0;
		return FALSE;
	}
	if (priv->move_dock_id) {
		win->autohide.state = KIBA_UNHIDDEN;
		priv->hide_id = 0;
		return FALSE;
	}
	allocation = dock->allocation;

	d = &kiba_layout_data[KIBA_DOCK(dock)->d.position];
	if (d->orientation == GTK_ORIENTATION_VERTICAL) {
		allocation.x += (d->halignment - 1)*win->autohide.speed;
		if (allocation.x < ((d->halignment-2)/2) * dock->allocation.width) {
			hidden = TRUE;
			allocation.x = (d->halignment - 1) * dock->allocation.width;
		}
		if (allocation.x > (d->halignment/2) * widget->allocation.width) {
			hidden = TRUE;
			allocation.x = (d->halignment/2) * widget->allocation.width;
		}
	} else {
		allocation.y += (d->valignment - 1)*win->autohide.speed;
		if (allocation.y < ((d->valignment-2)/2) * dock->allocation.height) {
			hidden = TRUE;
			allocation.y = ((d->valignment-2)/2) * allocation.height;
		}
		if (allocation.y > (d->valignment/2) * widget->allocation.height) {
			hidden = TRUE;
			allocation.y = (d->valignment/2) * widget->allocation.height;
		}
	}
	
	kiba_dock_queue_reposition(KIBA_DOCK(dock));
	gtk_widget_size_allocate (dock, &allocation);

	/* the dock is now hidden, exiting the timeout.. */
	if (hidden) {
		priv->hide_id = 0;
		win->autohide.state = KIBA_HIDDEN;
		kiba_dock_update_strut(dock);

		g_signal_emit (win,
			kiba_win_signals[SIG_KIBA_HIDDEN],
			0, dock);

		return FALSE;
	}

	return TRUE;
}

static gint
__unhide_dock (gpointer data)
{
	KibaWin *win = data;
	GtkWidget *widget = GTK_WIDGET(win);
	GtkWidget *dock = KIBA_WIN_CHILD(win);
	GtkAllocation allocation;
	KibaWinPrivate *priv = KIBA_WIN_GET_PRIVATE (win);
	const KibaLayoutData *d;
	gboolean unhidden = FALSE;

	if (priv->move_dock_id) {
		win->autohide.state = KIBA_UNHIDDEN;
		return FALSE;
	}

	allocation = dock->allocation;

	d = &kiba_layout_data[KIBA_DOCK(dock)->d.position];
	if (d->orientation == GTK_ORIENTATION_VERTICAL) {
		allocation.x -= (d->halignment - 1) * win->autohide.speed;

		if (d->halignment == 2 && allocation.x < widget->allocation.width - allocation.width) {
			unhidden = TRUE;
			allocation.y = widget->allocation.width - allocation.width;
		} else if (d->halignment == 0 && allocation.x > 0) {
			unhidden = TRUE;
			allocation.x = 0;
		}
	} else {
		allocation.y -= (d->valignment - 1) * win->autohide.speed;

		if (d->valignment == 2 && allocation.y < widget->allocation.height - allocation.height) {
			unhidden = TRUE;
			allocation.y = widget->allocation.height - allocation.height;
		} else if (d->valignment == 0 && allocation.y > 0) {
			unhidden = TRUE;
			allocation.y = 0;
		}
	}

	kiba_dock_queue_reposition(KIBA_DOCK(dock));
	gtk_widget_size_allocate (dock, &allocation);

	if (unhidden) {
		GList *l;
		
		for (l = GTK_BOX(dock)->children; l != NULL; l = l->next) {
			GtkBoxChild *child = l->data;
			kiba_object_make_sensitive(KIBA_OBJECT(child->widget), TRUE);
		}
		g_signal_emit (win,
			kiba_win_signals[SIG_KIBA_UNHIDDEN],
			0, dock);
	}

	return !unhidden;
}

static gint
__hide_delay (gpointer data)
{
	KibaWin *win = KIBA_WIN(data);
	KibaDock *dock = KIBA_WIN_DOCK(win);
	GList *l;
	
	win->autohide.state = KIBA_HIDING;
	for (l = GTK_BOX(dock)->children; l != NULL; l = l->next) {
		GtkBoxChild *child = l->data;
		kiba_object_make_sensitive(KIBA_OBJECT(child->widget), FALSE);
	}
	g_timeout_add (HIDE_RATE, __hide_dock, win);
	return FALSE;
}

static gint
__check_hide_dock (gpointer data)
{
	KibaWin *win = data;
	GtkWidget *widget = GTK_WIDGET(win);
	KibaWinPrivate  *priv;
	GtkWidget *dock = KIBA_WIN_CHILD(win);
	GdkRectangle monitor_size;
	gint x, y;
	gfloat area;
	
	gdk_screen_get_monitor_geometry (KIBA_GDK_SCREEN,
		kiba_screen_get_monitor(KIBA_SCREEN_DEFAULT), &monitor_size);
	gdk_window_get_pointer (dock->window, &x, &y, NULL);
	x -= dock->allocation.x - monitor_size.x;
	y -= dock->allocation.y - monitor_size.y;

	priv = KIBA_WIN_GET_PRIVATE (win);
	if (priv->move_dock_id)
		return TRUE;

	area = win->autohide.area;

	if (x > -area && x < dock->allocation.width + area &&
	    y > -area && y < dock->allocation.height + area) {
		if (win->autohide.state == KIBA_HIDING ||
		    win->autohide.state == KIBA_HIDDEN) {
			win->autohide.state = KIBA_UNHIDING;
			kiba_dock_update_strut(dock);
			g_timeout_add (HIDE_RATE , __unhide_dock, win);
		}
	} else if (!priv->hide_id && (win->autohide.state == KIBA_UNHIDING ||
		 win->autohide.state == KIBA_UNHIDDEN)) {
		priv->hide_id = g_timeout_add (win->autohide.delay*1000,
					__hide_delay, win);
	}

	return TRUE;
}

static gint
__raise_window (gpointer data)
{
	KibaWin *win = KIBA_WIN(data);
	KibaWinPrivate *priv = KIBA_WIN_GET_PRIVATE (win);

	gdk_window_set_keep_below (GTK_WIDGET (win)->window, TRUE);
	priv->raise_id = 0;

	return FALSE;
}

static GdkPixbuf*
get_desktop_background (KibaWin *win)
{
	GtkWidget *widget = GTK_WIDGET(win);
	Colormap cmap = GDK_COLORMAP_XCOLORMAP(
		gtk_widget_get_colormap(widget));
	Visual *visual = GDK_VISUAL_XVISUAL(
		gtk_widget_get_visual(widget));
	Display *display = GDK_DISPLAY();
	XID root_window = gdk_x11_drawable_get_xid(
		gdk_screen_get_root_window (KIBA_GDK_SCREEN));
	GdkPixbuf *pixbuf = NULL;
	gint format, i;
	Atom type;
	unsigned long nitems, bytes;
	unsigned int n;
	Pixmap *pixmap = 0;
	
	// Must init display
	gdk_pixbuf_xlib_init(display, 0);
	
	if (
			(
			 XGetWindowProperty(display, root_window, ATOM(_XROOTPMAP_ID), 0,
				 1, False, ATOM(PIXMAP), &type, &format,
				 &nitems, &bytes, (unsigned char**)&pixmap) == Success ||
			 XGetWindowProperty(display, root_window, ATOM(ESETROOT_PMAP_ID), 0,
				 1, False, ATOM(PIXMAP), &type, &format,
				 &nitems, &bytes, (unsigned char**)&pixmap) == Success
			)
			&&
			nitems == 1L && format == 32) {
		pixbuf = gdk_pixbuf_xlib_get_from_drawable (
			NULL, // destination
			*pixmap,
			cmap, visual, // colormap and visual, ignored for wins
			0, 0, // source offset
			0, 0, // destination offset
			KIBA_SCREEN_DEFAULT->geometry.width,
			KIBA_SCREEN_DEFAULT->geometry.height);
	}
	if (pixmap) {
		XFree(pixmap);
	}
	
	return pixbuf;
}

static void
draw_window_background(KibaWin *win, cairo_t *cr, gboolean before)
{
	cairo_set_operator (cr, CAIRO_OPERATOR_SOURCE);
	if (before) {
		cairo_set_operator (cr, CAIRO_OPERATOR_SOURCE);
		if (kiba_transparent_window_is_composited(
		    KIBA_TRANSPARENT_WINDOW(win))) {
			kiba_transparent_window_draw_background(
				KIBA_TRANSPARENT_WINDOW(win), cr);
		} else {
			cairo_set_source_rgba (cr, 0, 0, 0, 0);
			cairo_paint (cr);
		}
	} else {
		cairo_set_operator (cr, CAIRO_OPERATOR_DEST_OVER);
		if (!kiba_transparent_window_is_composited(
		    KIBA_TRANSPARENT_WINDOW(win))) {
			kiba_transparent_window_draw_background(
				KIBA_TRANSPARENT_WINDOW(win), cr);
		}
	}
}

/*
 * Clears the window to be transparent.
 */
static void
__clear (KibaWin *win, GtkAllocation allocation)
{
	KibaWinPrivate *priv = KIBA_WIN_GET_PRIVATE (win);
	GtkWidget *widget = GTK_WIDGET(win);
	cairo_t *cr;

	cr = kiba_win_cairo_create (win);
	if (cairo_status(cr) != CAIRO_STATUS_SUCCESS)
		return;
	
	draw_window_background(win, cr, TRUE);
	draw_window_background(win, cr, FALSE);
	cairo_destroy (cr);

	// renders the offscreen drawable to the window
#ifdef USE_GLITZ
	kiba_glitz_draw_func (priv->glitz_drawable,
		priv->glitz_draw_format);
#endif
#ifdef USE_OFFSCREEN_PIXMAP
	if (GDK_IS_PIXMAP(priv->pixmap)) {
		GdkGC *gc;

		gc = gdk_gc_new (priv->pixmap);
		if (GDK_IS_GC(gc)) {
			gdk_draw_drawable (widget->window,
				gc,
				priv->pixmap,
				0, 0, 0, 0,
				allocation.width,
				allocation.height);
			gdk_gc_unref (gc);
		}
	}
#endif
}

/*
 * Rerenders invalidate areas of the window.
 */
static gint
__redraw (gpointer data)
{
	GtkWidget *widget = data;
	KibaWinPrivate  *priv;
	KibaDock *dock;
	cairo_t *cr = NULL;
	GList *l;
	GdkGC *gc;
#ifdef USE_GLITZ
	// it seems to be better for the cpu usage to clip only one big rectangle for glitz
	GdkRectangle clip_box;
#endif
#ifndef USE_GLITZ
	GdkRectangle *invalid_rectangles = NULL;
	gint i, n_rectangles = 0;
#endif
	
	priv = KIBA_WIN_GET_PRIVATE (widget);
	dock = KIBA_WIN_DOCK(widget);
	
#ifdef USE_OFFSCREEN_PIXMAP
	if (priv->pixmap != NULL)
		g_object_unref(priv->pixmap);
	priv->pixmap = kiba_cairo_utils_create_pixmap (
				priv->choosen_allocation.width,
				priv->choosen_allocation.height,
				priv->depth);
	gc = gdk_gc_new (priv->pixmap);
#endif

	cr = kiba_win_cairo_create (KIBA_WIN(widget));
	if (cairo_status(cr) != CAIRO_STATUS_SUCCESS)
		return FALSE;

#if CLIPPING_CHECK == TRUE
	cairo_set_source_rgba (cr, 1, 0, 0, 1);
	cairo_set_operator (cr, CAIRO_OPERATOR_SOURCE);
	cairo_paint (cr);
#endif
	
	/* clip invalidate areas */
#ifndef USE_GLITZ
	gdk_region_get_rectangles (priv->invalid_region,
		&invalid_rectangles, &n_rectangles);
	for (i = 0; i < n_rectangles; i++) {
		GdkRectangle damaged_area = invalid_rectangles[i];
		cairo_rectangle (cr, damaged_area.x, damaged_area.y,
			damaged_area.width, damaged_area.height);
	}
#endif
#ifdef USE_OFFSCREEN_PIXMAP
#if CLIPPING_CHECK == FALSE
	gdk_gc_set_clip_region (gc, priv->invalid_region);
#endif
#endif
#ifdef USE_GLITZ
	gdk_region_get_clipbox (priv->invalid_region, &clip_box);
	cairo_rectangle (cr, clip_box.x, clip_box.y,
		clip_box.width, clip_box.height);
#endif
	cairo_clip (cr);

	priv->redraw_id = 0;
	gdk_region_empty(priv->invalid_region);

	draw_window_background(KIBA_WIN(widget), cr, TRUE);

	cairo_set_operator (cr, CAIRO_OPERATOR_OVER);

	if (KIBA_IS_DOCK(dock)) {
		// draw dock&children
		cairo_save (cr);
			kiba_dock_expose (dock, cr);
		cairo_restore (cr);
	}

	draw_window_background(KIBA_WIN(widget), cr, FALSE);
	
	cairo_destroy (cr);

	// rendering the offscreen drawable to the window
#ifdef USE_GLITZ
	kiba_glitz_draw_func (priv->glitz_drawable,
		priv->glitz_draw_format);
#endif
#ifdef USE_OFFSCREEN_PIXMAP
	gdk_draw_drawable (widget->window,
			   gc,
			   priv->pixmap,
			   0, 0, 0, 0,
			   priv->choosen_allocation.width,
			   priv->choosen_allocation.height);
	gdk_gc_unref (gc);
#endif

#if BENCHMARK_ENABLE == TRUE
	redraw_count++;
#endif

	return FALSE;
}

/*** GdkScreen signal handler ***/

static void
__composited_changed_cb(GdkScreen *screen,
		gpointer data)
{
	KibaWin *win = data;
	KibaWinPrivate *priv = KIBA_WIN_GET_PRIVATE (win);
	
	priv->composited = gdk_screen_is_composited (screen);
	
	if (!GDK_IS_WINDOW(GTK_WIDGET(win)->window)) {
		return;
	}
	
	if (priv->composited) {
		if (!GET_FLAG(win->probs, KIBA_WIN_BELOW)) {
			gdk_window_set_keep_below (GTK_WIDGET(win)->window, FALSE);
			gdk_window_raise(GTK_WIDGET(win)->window);
		}
		if (GET_FLAG(win->probs, KIBA_WIN_ABOVE)) {
			gdk_window_set_keep_above (GTK_WIDGET(win)->window, TRUE);
		}
	} else {
		gdk_window_set_keep_below (GTK_WIDGET(win)->window, TRUE);
		gdk_window_lower(GTK_WIDGET(win)->window);
	}
}

/*** KibaDock signal handler ***/

/*
 * Callback for the context menu items that offer different drag and drop receive methods.
 */
static void
__dnd_menu_item_activate_cb (GtkWidget *widget,
		gpointer data)
{
	KibaDragAction *dnd_action = data;
	dnd_action->data->receive_func (dnd_action->context,
			dnd_action->drag_data,
			dnd_action->data->user_data);
}

static void
__dnd_menu_destroy_cb (GtkWidget *widget,
		gpointer   data)
{
	GList *actions = data;
	GList *l;

	for (l = actions; l != NULL; l = actions) {
		KibaDragAction *dnd_action = l->data;

		actions = g_list_remove (actions, l->data);
		g_free (dnd_action->action);
		g_free (dnd_action->icon);
		g_free (dnd_action->drag_data);
		g_object_unref (G_OBJECT(dnd_action->context));
		g_free (dnd_action);
	}
	g_list_free (actions);
}

static void
__dock_drag_data_received_cb (GtkWidget *widget,
		GdkDragContext *drag_context,
		gint x,
		gint y,
		GtkSelectionData *selection_data,
		guint info,
		guint time,
		gpointer data)
{
	KibaWin *win = KIBA_WIN (data);
	KibaWinPrivate *priv = KIBA_WIN_GET_PRIVATE (win);
	GList *actions = NULL;
	GList *l;

	g_strstrip ((gchar *) selection_data->data);

	// checking for possible receive methods, adding them to actions list
	for (l = priv->drag_reveive_data; l != NULL; l = l->next) {
		KibaWinDragData *data = l->data;
		KibaDragAction *dnd_action;
		gchar *action = NULL, *icon = NULL;

		data->dest_func (drag_context,
				(gchar *) selection_data->data,
				data->user_data,
				&action, &icon);
		if (!action)
			continue;

		dnd_action = g_new0 (KibaDragAction, 1);
		dnd_action->data = data;
		dnd_action->action = g_strdup(action);
		dnd_action->icon = (icon != NULL ? g_strdup(icon) : "");
		dnd_action->drag_data = g_strdup ((gchar *) selection_data->data);
		dnd_action->context = g_object_ref (G_OBJECT(drag_context));

		actions = g_list_append (actions, dnd_action);
	}

	// displaying a menu to allow the user choosing the receive method
	if (g_list_length (actions) > 1) {
		GtkWidget *menu;
		GtkWidget *item;
		GtkWidget *image;

		menu = gtk_menu_new ();

		for (l = actions; l != NULL; l = l->next) {
			KibaDragAction *dnd_action = l->data;
			GdkPixbuf *pixbuf = NULL;

			item = gtk_image_menu_item_new_with_label (dnd_action->action);
			kiba_load_icon (dnd_action->icon, 24, &pixbuf);
			if (pixbuf) {
				image = gtk_image_new_from_pixbuf (pixbuf);
				g_object_unref (pixbuf);
			}
			else {
				image = gtk_image_new_from_stock (GTK_STOCK_MISSING_IMAGE,
						GTK_ICON_SIZE_MENU);
			}
			gtk_image_menu_item_set_image (GTK_IMAGE_MENU_ITEM (item), image);
			gtk_menu_shell_append (GTK_MENU_SHELL (menu), item);
			g_signal_connect (G_OBJECT (item), "activate",
				G_CALLBACK (__dnd_menu_item_activate_cb), dnd_action);
			gtk_widget_show (item);
		}

		gtk_menu_popup (GTK_MENU(menu),
			NULL, NULL, NULL, NULL,
			3, time);

		// connecting signal to free some memory
		g_signal_connect (G_OBJECT(menu), "destroy",
			G_CALLBACK (__dnd_menu_destroy_cb), actions);
	}
	// calling the one possible receive method
	else if (actions) {
		KibaDragAction *dnd_action = actions->data;

		dnd_action->data->receive_func (dnd_action->context,
				dnd_action->drag_data,
				dnd_action->data->user_data);

		actions = g_list_remove (actions, actions->data);
		g_free (dnd_action->action);
		g_free (dnd_action->icon);
		g_free (dnd_action->drag_data);
		g_object_unref (G_OBJECT(dnd_action->context));
		g_free (dnd_action);
		g_list_free (actions);
	}

	gtk_drag_finish (drag_context, TRUE, FALSE, time);
}

static gboolean
__dock_button_press_event (GtkWidget *widget,
		GdkEventButton *event,
		gpointer data)
{
	KibaWin *win  = data;
	KibaDock *dock = KIBA_DOCK (widget);
	KibaWinPrivate *priv = KIBA_WIN_GET_PRIVATE (win);
	GdkRectangle monitor_size;

	if (event->button == 3) {
		return FALSE;
	}
	
	gdk_screen_get_monitor_geometry (KIBA_GDK_SCREEN,
		kiba_screen_get_monitor(KIBA_SCREEN_DEFAULT), &monitor_size);
	gdk_window_get_pointer (
		gdk_screen_get_root_window(KIBA_GDK_SCREEN),
		&priv->mouse_anchor.x, &priv->mouse_anchor.y, NULL);
	priv->mouse_anchor.x -= monitor_size.x;
	priv->mouse_anchor.y -= monitor_size.y;
	
	// starting the dock drag
	if (!priv->move_dock_id) {
		if (win->autohide.state != KIBA_UNHIDDEN)
			win->autohide.state = KIBA_UNHIDDEN;
		priv->move_dock_id = g_timeout_add (20, __move_dock, win);
	}

	// setting drag alpha
	kiba_dock_set_alpha (KIBA_DOCK(widget), DOCK_DRAG_ALPHA);

	return FALSE;
}

static gboolean
__dock_button_release_event (GtkWidget *widget,
		GdkEventButton *event,
		gpointer data)
{
	KibaWin *win  = data;
	KibaDock *dock = KIBA_DOCK (widget);
	KibaWinPrivate *priv = KIBA_WIN_GET_PRIVATE (win);

	/* stopping the dock drag */
	if (priv->move_dock_id) {
		g_source_remove (priv->move_dock_id);
		priv->move_dock_id = 0;
	}

	priv->mouse_anchor.x = 0;
	priv->mouse_anchor.y = 0;

	/* reverting alpha */
	kiba_dock_set_alpha (dock, 1.0f);

	return FALSE;
}

static void
__dock_position_changed (KibaDock *dock,
		KibaPosition position,
		gpointer data)
{
	gtk_widget_queue_resize(GTK_WIDGET(data));
	kiba_dock_queue_update_surface (dock);
	// fixes artifacts on position change
	__clear (KIBA_WIN(data), GTK_WIDGET(data)->allocation);
}

static void
__dock_autohide_toggled (KibaDock  *dock,
		gboolean enabled,
		gpointer data)
{
	KibaWin *win = data;
	KibaWinPrivate  *priv = KIBA_WIN_GET_PRIVATE (win);

	if (enabled && !priv->check_hide_id) {
		priv->check_hide_id = g_timeout_add (CHECK_HIDE_RATE, __check_hide_dock, win);
	} else if (!enabled) {
		if (priv->check_hide_id) {
			g_source_remove (priv->check_hide_id);
			priv->check_hide_id = 0;
		}
		if (priv->hide_id) {
			g_source_remove (priv->hide_id);
			priv->hide_id = 0;
		}
		if (win->autohide.state == KIBA_HIDING || win->autohide.state == KIBA_HIDDEN) {
			win->autohide.state = KIBA_UNHIDING;
			g_timeout_add (HIDE_RATE , __unhide_dock, win);
		}
	}
}

/* GdkScreen Event */
static void
__screen_size_changed_cb (GdkScreen *screen, gpointer data)
{
	gtk_widget_queue_resize (GTK_WIDGET(data));
}

/*** Public Methods ***/

/**
 * kiba_win_add_drag_func:
 * @win: The KibaWin Instance.
 * @receive_func: The method to receive drag sources.
 * @dest_func: The method to determine if a drag source can be received.
 * @user_data: The user data.
 *
 * Adds drag and drop receive function.
 */
void kiba_win_add_drag_func (KibaWin *win,
		KibaDragReceiveFunc receive_func,
		KibaDragDestFunc dest_func,
		gpointer user_data)
{
	KibaWinPrivate *priv;
	KibaWinDragData *new_data;
	GList *l;

	g_return_if_fail (KIBA_IS_WIN(win));
	g_return_if_fail (receive_func != NULL);
	g_return_if_fail (dest_func != NULL);

	priv = KIBA_WIN_GET_PRIVATE (win);

	for (l = priv->drag_reveive_data; l != NULL; l = l->next) {
		KibaWinDragData *data = l->data;

		if (data->receive_func == receive_func) {
			data->user_data = user_data;
			return;
		}
	}

	new_data = g_new0 (KibaWinDragData, 1);
	new_data->receive_func = receive_func;
	new_data->dest_func    = dest_func;
	new_data->user_data    = user_data;

	priv->drag_reveive_data = g_list_append (priv->drag_reveive_data, new_data);
}

/**
 * kiba_win_remove_drag_func:
 * @win: The KibaWin Instance.
 * @receive_func: The drag and drop receive method to remove.
 *
 * Removes previously added drag and drop receive function.
 */
void
kiba_win_remove_drag_func (KibaWin *win,
	KibaDragReceiveFunc receive_func)
{
	g_return_if_fail (KIBA_IS_WIN(win));
	
	KibaWinPrivate *priv = KIBA_WIN_GET_PRIVATE (win);
	GList *l;

	for (l = priv->drag_reveive_data; l != NULL; l = l->next) {
		KibaWinDragData *data = l->data;

		if (data->receive_func == receive_func) {
			priv->drag_reveive_data = g_list_remove (priv->drag_reveive_data, data);
			g_free (data);
			return;
		}
	}
}

/**
 * kiba_win_set_flag:
 * @win: The KibaWin Instance.
 * @prob: The property to set.
 *
 * Sets a property.
 */
void
kiba_win_set_flag (KibaWin *win,
		KibaWinProbs prob)
{
	g_return_if_fail (KIBA_IS_WIN(win));
	
	KibaWinPrivate *priv = KIBA_WIN_GET_PRIVATE (win);
	GtkWidget *child = gtk_bin_get_child(GTK_BIN(win));

	if (GET_FLAG(win->probs, prob))
		return;

	SET_FLAG(win->probs, prob);

	if (!GTK_WIDGET_REALIZED(win))
		return;

	switch (prob) {
		case KIBA_WIN_ABOVE:
			if (priv->composited) {
				gdk_window_set_keep_above (GTK_WIDGET (win)->window, TRUE);
			}
			break;
		case KIBA_WIN_BELOW:
			if (priv->composited) {
				gdk_window_set_keep_below (GTK_WIDGET (win)->window, TRUE);
			}
			break;
		default:
			g_assert_not_reached();
			break;
	}
}

/**
 * kiba_win_unset_flag:
 * @win: The KibaWin Instance.
 * @prob: The property to unset.
 *
 * Unsets a property.
 */
void
kiba_win_unset_flag (KibaWin *win,
		KibaWinProbs prob)
{
	g_return_if_fail (KIBA_IS_WIN(win));
	
	KibaWinPrivate *priv = KIBA_WIN_GET_PRIVATE (win);
	GtkWidget *child = gtk_bin_get_child(GTK_BIN(win));
	
	if (!GET_FLAG(win->probs, prob))
		return;

	DEL_FLAG(win->probs, prob);
	if (!GTK_WIDGET_REALIZED(win))
		return;

	switch (prob) {
		case KIBA_WIN_ABOVE:
			if (priv->composited) {
				gdk_window_set_keep_above (GTK_WIDGET(win)->window, FALSE);
			}
			break;
		case KIBA_WIN_BELOW:
			if (priv->composited) {
				gdk_window_set_keep_below (GTK_WIDGET (win)->window, FALSE);
				if (child)
					kiba_dock_update_strut(child);
			}
			break;
		default:
			g_assert_not_reached();
			break;
	}
}

/**
 * kiba_win_invalidate_area:
 * @widget: The KibaWin Instance.
 * @area: The area that should be redrawn on next expose event.
 *
 * Invalidate area on win and queues draw.
 */
void
kiba_win_invalidate_area (GtkWidget *widget,
		GdkRectangle area)
{
	g_return_if_fail (KIBA_IS_WIN(widget));
	
	KibaWinPrivate  *priv = KIBA_WIN_GET_PRIVATE (widget);

	if (area.width <= 0 || area.height <= 0) {
		DEBUG_WARNING("Cannot invalidate area: x=%i,x=%i,w=%i,h=%i\n",
			area.x, area.y,
			area.width, area.height);
		return;
	}
	
	// avoid off-window areas
	if (area.x < 0) {
		area.x = 0;
	} if (area.x + area.width > priv->choosen_allocation.width) {
		area.width = priv->choosen_allocation.width - area.x;
	}
	if (area.y < 0) {
		area.y = 0;
	} if (area.y + area.height > priv->choosen_allocation.height) {
		area.height = priv->choosen_allocation.height - area.y;
	}
	
	if (!priv->redraw_id) {
		priv->redraw_id = g_timeout_add (REDRAW_RATE, __redraw, widget);
	}
	
	gdk_region_union_with_rect (priv->invalid_region, &area);
}

/**
 * kiba_win_cairo_create:
 * @win: The KibaWin Instance.
 *
 * Creates a cairo context, compatible to the KibaWin window.
 *
 * Returns: the drawing context created.
 */
cairo_t*
kiba_win_cairo_create (KibaWin *win)
{
	g_return_val_if_fail (KIBA_IS_WIN(win), NULL);

	if (!GTK_WIDGET_REALIZED(win))
		return NULL;
	
	KibaWinPrivate  *priv = KIBA_WIN_GET_PRIVATE (win);
	GtkWidget *widget = GTK_WIDGET(win);
	cairo_t *cr;
	
#ifdef USE_GLITZ
	cr = kiba_glitz_get_cairo_context (priv->glitz_drawable,
			priv->glitz_draw_format,
			priv->glitz_format,
			priv->choosen_allocation.width,
			priv->choosen_allocation.height);
	if (cairo_status(cr) == CAIRO_STATUS_SUCCESS)
		return cr;
#endif
#ifdef USE_OFFSCREEN_PIXMAP
	cr = gdk_cairo_create (priv->pixmap);
	if (cairo_status(cr) == CAIRO_STATUS_SUCCESS)
		return cr;
#endif
	cairo_destroy (cr);

	return gdk_cairo_create (widget->window);
}

/**
 * kiba_win_quit:
 * @widget:  The KibaWin Instance.
 *
 * Hide the dock.
 *
 * Returns: FALSE if the dock is hidden, otherwise TRUE is returned.
 */
gboolean
kiba_win_quit (GtkWidget *widget)
{
	KibaWin *win = KIBA_WIN(widget);
	KibaWinPrivate *priv;
	GtkWidget *dock;

	g_return_if_fail (KIBA_IS_WIN(widget));

	priv = KIBA_WIN_GET_PRIVATE (widget);
	dock = GTK_BIN(widget)->child;

	if (priv->check_hide_id) {
		g_source_remove (priv->check_hide_id);
		priv->check_hide_id = 0;
	}
	win->autohide.state = KIBA_HIDING;

	if (!__hide_dock(widget))
		return FALSE;
	
	if (priv->invalid_region) {
		kiba_win_invalidate_area (widget, (GdkRectangle) {
			0, 0,
			priv->choosen_allocation.width,
			priv->choosen_allocation.height});
		__redraw (widget);
	}
	
	return TRUE;
}

/**
 * kiba_win_add_visual_feature:
 * @win: The KibaWin Instance.
 * @feature: The KibaVisualFeature instance.
 *
 * Returns: TRUE on success
 */
gboolean
kiba_win_add_visual_feature(KibaWin *win, KibaVisualFeature *feature)
{
	g_return_val_if_fail(KIBA_IS_WIN(win), FALSE);
	g_return_val_if_fail(KIBA_IS_VISUAL_FEATURE(feature), FALSE);
	
	KibaWinPrivate *priv = KIBA_WIN_GET_PRIVATE (win);
	priv->features = g_list_prepend(priv->features, feature);
}

/**
 * kiba_win_remove_visual_feature:
 * @win: The KibaWin Instance.
 * @name: The KibaVisualFeature name.
 *
 * Returns: TRUE on success
 */
gboolean
kiba_win_remove_visual_feature(KibaWin *win, const gchar *name)
{
	g_return_val_if_fail(KIBA_IS_WIN(win), FALSE);
	g_return_val_if_fail(name != NULL, FALSE);
	
	KibaWinPrivate *priv = KIBA_WIN_GET_PRIVATE (win);
	KibaVisualFeature *feature = kiba_win_get_feature_by_name(win, name);
	
	if (feature) {
		priv->features = g_list_remove(priv->features, feature);
		g_object_unref(G_OBJECT(feature));
		return TRUE;
	} else {
		return FALSE;
	}
}

/**
 * kiba_win_get_feature_by_name:
 * @win: The KibaWin Instance.
 * @name: The KibaVisualFeature name.
 *
 * Returns: The KibaVisualFeature instance found or NULL
 */
KibaVisualFeature*
kiba_win_get_feature_by_name(KibaWin *win, const gchar *name)
{
	g_return_val_if_fail(KIBA_IS_WIN(win), NULL);
	g_return_val_if_fail(name != NULL, NULL);
	
	KibaWinPrivate *priv = KIBA_WIN_GET_PRIVATE (win);
	GList *l;
	
	for (l = priv->features; l != NULL; l = l->next) {
		KibaVisualFeature *feature = l->data;
		if (strcmp(name, kiba_visual_feature_get_name(feature)) == 0) {
			return feature;
		}
	}
	
	return NULL;
}

/*** preferences callback ***/

static void
preferences_changed_cb (const gchar *option_id,
		gpointer value,
		gpointer data)
{
	KibaWin *win = data;
	
	if (strcmp(option_id, PREFERENCES_BELOW_KEY) == 0) {
		if ((gboolean) GPOINTER_TO_INT(value) == TRUE) {
			kiba_win_set_flag (win, KIBA_WIN_BELOW);
		} else {
			kiba_win_unset_flag (win, KIBA_WIN_BELOW);
		}
	} else if (strcmp(option_id, PREFERENCES_ABOVE_KEY) == 0) {
		if ((gboolean) GPOINTER_TO_INT(value) == TRUE) {
			kiba_win_set_flag (win, KIBA_WIN_ABOVE);
		} else {
			kiba_win_unset_flag (win, KIBA_WIN_ABOVE);
		}
	} else if (strcmp(option_id, PREFERENCES_HIDE_DELAY_KEY) == 0) {
		win->autohide.delay = GPOINTER_TO_INT (value);
	} else if (strcmp(option_id, PREFERENCES_HIDE_SPEED_KEY) == 0) {
		win->autohide.speed = GPOINTER_TO_INT (value);
	} else if (strcmp(option_id, PREFERENCES_HIDE_AREA_KEY) == 0) {
		win->autohide.area = GPOINTER_TO_INT (value);
	} else if (strcmp(option_id, PREFERENCES_EDGE_OFFSET_KEY) == 0) {
		win->offset = GPOINTER_TO_INT(value);
		gtk_widget_queue_resize_no_redraw(GTK_WIDGET(win));
	} else {
		DEBUG_ERROR("Unknow KibaWin option: %s\n", option_id);
	}
}

/*** Widget Class Implementations ***/

/*
 * gtk_container_remove implementation.
 */
static void
kiba_win_remove (GtkContainer *container,
		GtkWidget *widget)
{
	KibaWinPrivate *priv = KIBA_WIN_GET_PRIVATE (container);

	// disconnecting the KibaDock event handler
	g_signal_handlers_disconnect_by_func (G_OBJECT(widget),
		G_CALLBACK(__dock_button_press_event), container);
	g_signal_handlers_disconnect_by_func (G_OBJECT(widget),
		G_CALLBACK(__dock_button_release_event), container);
	g_signal_handlers_disconnect_by_func (G_OBJECT(widget),
		G_CALLBACK(__dock_position_changed), container);
	g_signal_handlers_disconnect_by_func (G_OBJECT(widget),
		G_CALLBACK(__dock_autohide_toggled), container);
	g_signal_handlers_disconnect_by_func (G_OBJECT(widget),
		G_CALLBACK(__dock_drag_data_received_cb), container);

	if (priv->check_hide_id) {
		g_source_remove (priv->check_hide_id);
		priv->check_hide_id = 0;
	}

	(* GTK_CONTAINER_CLASS (kiba_win_parent_class)->remove) (container, widget);
}

/*
 * gtk_container_add implementation.
 */
static void
kiba_win_add (GtkContainer *container,
		GtkWidget *widget)
{
	KibaWinPrivate *priv = KIBA_WIN_GET_PRIVATE (container);

	g_return_if_fail (KIBA_IS_DOCK(widget));

	// connecting to KibaDock events
	g_signal_connect_after (G_OBJECT(widget), "button-press-event",
		G_CALLBACK(__dock_button_press_event), container);
	g_signal_connect (G_OBJECT(widget), "button-release-event",
		G_CALLBACK(__dock_button_release_event), container);
	g_signal_connect (G_OBJECT(widget), "position-changed",
		G_CALLBACK(__dock_position_changed), container);
	g_signal_connect (G_OBJECT(widget), "autohide-toggled",
		G_CALLBACK(__dock_autohide_toggled), container);
	g_signal_connect_after (G_OBJECT(widget), "drag-data-received",
		G_CALLBACK(__dock_drag_data_received_cb), container);

	if (GET_FLAG(KIBA_DOCK(widget)->flags, KIBA_DOCK_AUTOHIDE) && !priv->check_hide_id)
		priv->check_hide_id = g_timeout_add (CHECK_HIDE_RATE, __check_hide_dock, container);

	(* GTK_CONTAINER_CLASS (kiba_win_parent_class)->add) (container, widget);
}

/*
 * gtk_widget_expose implementation.
 */
static gboolean
kiba_win_expose_event (GtkWidget *widget,
	  GdkEventExpose *event)
{
	KibaWinPrivate  *priv;

	if (!GTK_WIDGET_REALIZED (widget))
		return FALSE;

	priv = KIBA_WIN_GET_PRIVATE (widget);

	kiba_win_invalidate_area (widget, event->area);

	// using maximal MAX_FPS fps
	if (!priv->redraw_id)
		priv->redraw_id = g_timeout_add (REDRAW_RATE, __redraw, widget);

	return FALSE;
}

/*
 * gtk_widget_screen_changed implementation.
 */
static void
kiba_win_screen_changed (GtkWidget *widget,
		GdkScreen *previous_screen)
{
	KibaWinPrivate *priv = KIBA_WIN_GET_PRIVATE (widget);
	GdkScreen *screen = gtk_widget_get_screen (widget);
	GdkColormap *colormap = NULL;
	GdkVisual *visual = NULL;
	
	// update colormap and visual
#ifdef USE_GLITZ
	kiba_glitz_get_colormap (&priv->glitz_draw_format,
		&visual,
		&colormap);
#endif
	if (!colormap || !visual) {
		colormap = gdk_screen_get_rgba_colormap (screen);
		visual = gdk_screen_get_rgba_visual (screen);
	}
	if (!colormap || !visual) {
		colormap = gdk_screen_get_rgb_colormap (screen);
		visual   = gdk_screen_get_rgb_visual (screen);
	}
	gtk_widget_set_visual (widget, visual);
	gtk_widget_set_colormap (widget, colormap);
	
	// keep track of screen size
	if (GDK_IS_SCREEN(previous_screen)) {
		g_signal_handlers_disconnect_by_func (G_OBJECT(previous_screen),
			G_CALLBACK(__screen_size_changed_cb), widget);
	}
	g_signal_connect (G_OBJECT(screen), "size-changed",
		G_CALLBACK(__screen_size_changed_cb), widget);
    
	// keep track of composite manager
	if (GDK_IS_SCREEN(previous_screen)) {
		g_signal_handlers_disconnect_by_func (G_OBJECT(previous_screen),
			G_CALLBACK(__composited_changed_cb), widget);
	}
	g_signal_connect (G_OBJECT(screen), "composited-changed",
		G_CALLBACK(__composited_changed_cb), widget);
	__composited_changed_cb(screen, widget);
}

/*
 * gtk_widget_size_allocate implementation.
 */
static void
kiba_win_size_allocate (GtkWidget  *widget,
		GtkAllocation *allocation)
{
	KibaWinPrivate  *priv = KIBA_WIN_GET_PRIVATE (widget);
	KibaWin *win = KIBA_WIN(widget);
	GtkWidget *dock = KIBA_WIN_CHILD(widget);
	gboolean size_changed, position_changed;
	GtkAllocation old_alloc;
	gint ww = 0, wh = 0;
	GList *l;
#ifdef USE_OFFSCREEN_PIXMAP
	GdkGC *gc;
#endif
	
	// FIXME: allocation is set somewhere, must be gtk :/
	widget->allocation = priv->last_allocation;
	
	size_changed = (priv->choosen_allocation.width != widget->allocation.width
		|| priv->choosen_allocation.height != widget->allocation.height);
	position_changed = (allocation->x != priv->choosen_allocation.x ||
		allocation->y != priv->choosen_allocation.y);
	
	old_alloc = widget->allocation;
	allocation->x = priv->choosen_allocation.x;
	allocation->y = priv->choosen_allocation.y;
	widget->allocation = *allocation;
	priv->last_allocation = widget->allocation;
	if (!GTK_WIDGET_REALIZED(widget)) {
		return;
	}
	
	if (dock) {
		__resize_children (win);
	};
	
	if (!size_changed) {
		if (position_changed && !priv->composited) {
			// redraw background on position change
			kiba_win_invalidate_area (widget, (GdkRectangle) {
				0, 0,
				priv->choosen_allocation.width,
				priv->choosen_allocation.height});
		}

		gdk_window_move(widget->window,
			priv->choosen_allocation.x,
			priv->choosen_allocation.y);
		return;
	}
	if (dock) {
		kiba_dock_queue_update_surface (KIBA_DOCK(dock));
	};
	
	
	// update offscreen surface
#ifdef USE_GLITZ
	kiba_glitz_resize_func (priv->glitz_drawable,
		priv->choosen_allocation.width,
		priv->choosen_allocation.height);
#endif
#ifdef USE_OFFSCREEN_PIXMAP
	if (GDK_IS_PIXMAP(priv->pixmap))
		g_object_unref(priv->pixmap);
	priv->pixmap = kiba_cairo_utils_create_pixmap (
		priv->choosen_allocation.width,
		priv->choosen_allocation.height, priv->depth);
	gc = gdk_gc_new (priv->pixmap);
#endif
	
	cairo_t *cr = NULL;
	// draw dock to the new surface
	cr = kiba_win_cairo_create (KIBA_WIN(widget));
	if (cairo_status(cr) == CAIRO_STATUS_SUCCESS) {
		draw_window_background(KIBA_WIN(widget), cr, TRUE);
		
		cairo_set_operator (cr, CAIRO_OPERATOR_OVER);
		cairo_save (cr);
			kiba_dock_expose (KIBA_DOCK(dock), cr);
		cairo_restore (cr);
		
		draw_window_background(KIBA_WIN(widget), cr, FALSE);

		cairo_destroy (cr);
	}
	
	// move and resize window
	gdk_window_move_resize(widget->window,
		priv->choosen_allocation.x, priv->choosen_allocation.y,
		priv->choosen_allocation.width, priv->choosen_allocation.height);
	gdk_window_configure_finished (widget->window);
	
	// draw onto resized window
#ifdef USE_GLITZ
	kiba_glitz_draw_func (priv->glitz_drawable, priv->glitz_draw_format);
#endif
#ifdef USE_OFFSCREEN_PIXMAP
	gdk_draw_drawable (widget->window,
		gc, priv->pixmap,
		0, 0, 0, 0,
		priv->choosen_allocation.width,
		priv->choosen_allocation.height);
	gdk_gc_unref (gc);
#endif
}

/*
 * gtk_widget_size_request implementation.
 */
static void
kiba_win_size_request (GtkWidget *widget,
		GtkRequisition *requisition)
{
	GtkWidget *dock = KIBA_WIN_CHILD(widget);
	KibaWinPrivate *priv = KIBA_WIN_GET_PRIVATE (widget);

	if (!GTK_WIDGET_REALIZED(widget)) {
		requisition->width = 0;
		requisition->height = 0;
	} else {
		gtk_widget_size_request (dock, requisition);
		
		// fit dock onto edge with respect to other docks
		__choose_allocation(KIBA_WIN(widget), requisition);
		requisition->height = priv->choosen_allocation.height;
		requisition->width = priv->choosen_allocation.width;
	}
}

/*
 * gtk_widget_enter_notify_event implementation.
 */
static gboolean
kiba_win_enter_notify_event (GtkWidget *widget,
		GdkEventCrossing *event)
{
	KibaWin *win = KIBA_WIN(widget);
	KibaWinPrivate *priv = KIBA_WIN_GET_PRIVATE (widget);

	if (priv->composited && GET_FLAG(win->probs, KIBA_WIN_BELOW)) {
		// raise window on enter event
		if (priv->raise_id) {
			g_source_remove(priv->raise_id);
			priv->raise_id = 0;
		}
		gdk_window_set_keep_below (GTK_WIDGET (win)->window, FALSE);
	}

	return FALSE;
}

/*
 * gtk_widget_leave_notify_event implementation.
 */
static gboolean
kiba_win_leave_notify_event (GtkWidget *widget,
		GdkEventCrossing *event)
{
	KibaWin *win = KIBA_WIN(widget);
	KibaWinPrivate *priv = KIBA_WIN_GET_PRIVATE (widget);
	GtkWidget *dock;

	dock = KIBA_WIN_CHILD(win);

	if (priv->composited && GET_FLAG(win->probs, KIBA_WIN_BELOW)) {
		// FIXME: there can be leave events from childs with own windows
		if (!priv->raise_id)
			priv->raise_id = g_timeout_add (500, __raise_window, win);
	}

	return FALSE;
}

/*
 * gtk_widget_show implementation.
 */
static void
kiba_win_show (GtkWidget *widget)
{
	GTK_WIDGET_SET_FLAGS (widget, GTK_VISIBLE);

	if (!GTK_WIDGET_REALIZED (widget))
		gtk_widget_realize (widget);

	if (!GTK_WIDGET_MAPPED (widget))
		gtk_widget_map (widget);

	/* clear the window to be transparent */
	__clear (KIBA_WIN(widget), widget->allocation);
}

/*
 * gtk_widget_map implementation.
 */
static void
kiba_win_map (GtkWidget *widget)
{
	GtkBin *bin = GTK_BIN (widget);

	GTK_WIDGET_SET_FLAGS (widget, GTK_MAPPED);

	gdk_window_show (widget->window);

	if (bin->child && GTK_WIDGET_VISIBLE (bin->child) && !GTK_WIDGET_MAPPED (bin->child))
		gtk_widget_map (bin->child);
}

/*
 * gtk_widget_unrealize implementation.
 */
static void
kiba_win_unrealize (GtkWidget *widget)
{
	KibaWinPrivate *priv = KIBA_WIN_GET_PRIVATE (widget);
	KibaWin *win = KIBA_WIN(widget);
	GList *l;
	
	(* GTK_WIDGET_CLASS (kiba_win_parent_class)->unrealize) (widget);
	
	g_signal_handlers_disconnect_by_func (G_OBJECT(KIBA_GDK_SCREEN),
		G_CALLBACK(__screen_size_changed_cb), widget);
	g_signal_handlers_disconnect_by_func (G_OBJECT(KIBA_GDK_SCREEN),
		G_CALLBACK(__composited_changed_cb), widget);

	for (l = priv->drag_reveive_data; l != NULL; l = priv->drag_reveive_data) {
		KibaWinDragData *data = l->data;
		priv->drag_reveive_data = g_list_remove (priv->drag_reveive_data, l->data);
		g_free (data);
	}
	g_list_free (priv->drag_reveive_data);
	priv->drag_reveive_data = NULL;
	
	gdk_region_destroy(priv->invalid_region);
	priv->invalid_region = NULL;

	if (priv->raise_id)
		g_source_remove (priv->raise_id);
	priv->raise_id = 0;
	if (priv->redraw_id)
		g_source_remove (priv->redraw_id);
	priv->redraw_id = 0;
	if (priv->check_hide_id)
		g_source_remove (priv->check_hide_id);
	priv->check_hide_id = 0;
	if (priv->hide_id)
		g_source_remove (priv->hide_id);
	priv->hide_id = 0;
	if (priv->move_dock_id)
		g_source_remove (priv->move_dock_id);
	priv->move_dock_id = 0;

#ifdef USE_GLITZ
	kiba_glitz_fini (priv->glitz_drawable);
	priv->glitz_drawable = NULL;
#endif
#ifdef USE_OFFSCREEN_PIXMAP
	if (GDK_IS_PIXMAP(priv->pixmap)) {
		g_object_unref (priv->pixmap);
		priv->pixmap = NULL;
	}
#endif
}

/*
 * gtk_widget_realize implementation.
 */
static void
kiba_win_realize (GtkWidget *widget)
{
	GtkWindow *window = GTK_WINDOW (widget);
	KibaWinPrivate *priv = KIBA_WIN_GET_PRIVATE (widget);
	KibaWin *win = KIBA_WIN(widget);
	GdkWindowAttr attributes;
	gint attributes_mask;
	
	// make sure the size is properly allocated
	if (widget->allocation.x == -1 &&
	    widget->allocation.y == -1 &&
	    widget->allocation.width == 1 &&
	    widget->allocation.height == 1) {
		GtkRequisition requisition;
		GtkAllocation allocation = { 0, 0, 1, 1 };
		priv->choosen_allocation = allocation;
		
		gtk_widget_size_request (widget, &requisition);
		allocation.width = requisition.width;
		allocation.height = requisition.height;
		gtk_widget_size_allocate (widget, &allocation);
	}

	GTK_WIDGET_SET_FLAGS (widget, GTK_REALIZED);

	attributes.window_type = GDK_WINDOW_TOPLEVEL;
	attributes.wclass = GDK_INPUT_OUTPUT;
	attributes.title = window->title;
	attributes.wmclass_name = window->wmclass_name;
	attributes.wmclass_class = window->wmclass_class;
	attributes.width = widget->allocation.width;
	attributes.height = widget->allocation.height;
	attributes.x = widget->allocation.x;
	attributes.y = widget->allocation.y;
	attributes.visual = gtk_widget_get_visual(widget);
	attributes.colormap = gtk_widget_get_colormap(widget);
	attributes.event_mask = gtk_widget_get_events (widget) | GDK_EXPOSURE_MASK;
	attributes_mask = GDK_WA_X | GDK_WA_Y;
	if (attributes.visual != NULL && attributes.colormap != NULL)
		attributes_mask |= GDK_WA_VISUAL | GDK_WA_COLORMAP;
	attributes_mask |= (window->title ? GDK_WA_TITLE : 0);
	attributes_mask |= (window->wmclass_name ? GDK_WA_WMCLASS : 0);

	widget->window = gdk_window_new (gdk_get_default_root_window (),
				&attributes, attributes_mask);
	// synchronizing the window repaint with the window manager
	// during resizing operations
	gdk_window_enable_synchronized_configure (widget->window);
	// catch the window events on the widget
	gdk_window_set_user_data (widget->window, window);
	widget->style = gtk_style_attach (widget->style, widget->window);

	gdk_window_set_type_hint (widget->window,
		gtk_window_get_type_hint(window));
	gdk_window_set_icon_name (widget->window, PACKAGE);
	gdk_window_stick (widget->window);
	
	if (priv->composited) {
		if (GET_FLAG(KIBA_WIN(widget)->probs, KIBA_WIN_ABOVE)) {
			gdk_window_set_keep_above (widget->window, TRUE);
		}
		if (GET_FLAG(KIBA_WIN(widget)->probs, KIBA_WIN_BELOW)) {
			gdk_window_set_keep_below (widget->window, TRUE);
			// without that the window has border !?!
			gdk_window_set_decorations (widget->window, 0);
		}
	} else {
		gdk_window_set_keep_above (widget->window, FALSE);
		gdk_window_set_keep_below (widget->window, TRUE);
	}
	priv->depth = attributes.visual->depth;

	priv->invalid_region = gdk_region_new();
	
	// initialling the offscreen drawable, the double buffering is disabled
#ifdef USE_GLITZ
	kiba_glitz_create_drawable (GDK_WINDOW_XWINDOW(widget->window),
		widget->allocation.width, widget->allocation.height,
		priv->glitz_draw_format,
		&priv->glitz_format,
		&priv->glitz_drawable);
#endif
#ifdef USE_OFFSCREEN_PIXMAP
	priv->pixmap = kiba_cairo_utils_create_pixmap (widget->allocation.width,
			widget->allocation.height, priv->depth);
#endif

#if BENCHMARK_ENABLE == TRUE
	// spitting out redraws
	g_timeout_add (1000, __benchmark_cb, NULL);
#endif
	
	kiba_greyed_feature_new(win);
}

/*** Class init ***/

static void
kiba_win_class_init (KibaWinClass *class)
{
	GtkWidgetClass *widget_class = GTK_WIDGET_CLASS (class);
	GtkContainerClass *container_class = GTK_CONTAINER_CLASS (class);

	g_type_class_add_private (class, sizeof (KibaWinPrivate));

	widget_class->realize = kiba_win_realize;
	widget_class->unrealize = kiba_win_unrealize;
	widget_class->show = kiba_win_show;
	widget_class->map = kiba_win_map;
	widget_class->enter_notify_event = kiba_win_enter_notify_event;
	widget_class->leave_notify_event = kiba_win_leave_notify_event;
	widget_class->size_allocate = kiba_win_size_allocate;
	widget_class->size_request = kiba_win_size_request;
	widget_class->expose_event = kiba_win_expose_event;
	widget_class->screen_changed = kiba_win_screen_changed;

	container_class->add = kiba_win_add;
	container_class->remove = kiba_win_remove;

	// signals
	kiba_win_signals[SIG_KIBA_HIDDEN] =
		g_signal_new ("dock-hidden",
			G_OBJECT_CLASS_TYPE (class),
			G_SIGNAL_RUN_LAST,
			G_STRUCT_OFFSET (KibaWinClass, dock_hidden),
			NULL, NULL,
			g_cclosure_marshal_VOID__OBJECT,
			G_TYPE_NONE, 1,
			KIBA_TYPE_DOCK);
	kiba_win_signals[SIG_KIBA_UNHIDDEN] =
		g_signal_new ("dock-unhidden",
			G_OBJECT_CLASS_TYPE (class),
			G_SIGNAL_RUN_LAST,
			G_STRUCT_OFFSET (KibaWinClass, dock_unhidden),
			NULL, NULL,
			g_cclosure_marshal_VOID__OBJECT,
			G_TYPE_NONE, 1,
			KIBA_TYPE_DOCK);
}

static void
kiba_win_init (KibaWin *kiba_win)
{
	KibaWinPrivate *priv;
	KibaScreen *screen;

	GTK_WIDGET_UNSET_FLAGS (GTK_WIDGET (kiba_win),
			GTK_NO_WINDOW |
			GTK_DOUBLE_BUFFERED);

	priv = KIBA_WIN_GET_PRIVATE (kiba_win);
	screen = kiba_screen_get();

	gtk_widget_add_events (GTK_WIDGET (kiba_win),
		GDK_EXPOSURE_MASK |
		GDK_ENTER_NOTIFY_MASK |
		GDK_LEAVE_NOTIFY_MASK |
		GDK_POINTER_MOTION_MASK |
		GDK_POINTER_MOTION_HINT_MASK);
	gtk_widget_set_redraw_on_allocate (
		GTK_WIDGET(kiba_win), FALSE);
	
	kiba_win_screen_changed (GTK_WIDGET (kiba_win), NULL);
	gtk_widget_set_app_paintable (GTK_WIDGET (kiba_win), TRUE);
	gtk_window_set_decorated (GTK_WINDOW (kiba_win), FALSE);
	gtk_window_set_resizable (GTK_WINDOW (kiba_win), FALSE);
	gtk_window_set_title (GTK_WINDOW (kiba_win), PACKAGE);
	gtk_window_set_type_hint (GTK_WINDOW(kiba_win), GDK_WINDOW_TYPE_HINT_DOCK);
	gtk_window_stick (GTK_WINDOW(kiba_win));

	kiba_win->autohide.state = KIBA_UNHIDDEN;
	kiba_win->autohide.area = 25;
	kiba_win->autohide.speed = 3;
	kiba_win->autohide.delay = 2;
	
	priv->invalid_region = NULL;
	priv->check_hide_id = 0;
	priv->redraw_id = 0;
	priv->move_dock_id = 0;
	priv->hide_id = 0;
	priv->mouse_anchor = (GdkRectangle) { 0, 0, 0, 0 };
	priv->drag_reveive_data = NULL;
#ifdef USE_GLITZ
	priv->glitz_drawable = NULL;
	priv->glitz_format = NULL;
#endif
#ifdef USE_OFFSCREEN_PIXMAP
	priv->pixmap = NULL;
#endif
}

/**
 * kiba_win_new:
 * @configFile: configuration file associated to the window.
 *
 * Creates and returns a KibaWin Instance.
 *
 * Returns: the KibaWin Instance created.
 */
GtkWidget* kiba_win_new (KibaPrefsFile *configFile)
{
	GtkWidget *win = g_object_new (KIBA_TYPE_WIN, NULL);
	gint i;
	
	// add preference callbacks
	kiba_prefs_add_options (configFile,
		win_options, G_N_ELEMENTS(win_options));
	for (i = 0; i < G_N_ELEMENTS(win_options); i++) {
		kiba_prefs_add_callback (configFile, win_options[i].id,
			preferences_changed_cb, win, TRUE);
	}
	
	return win;
}
